"""Tests for runway.cfngin.actions.base."""
# pylint: disable=no-self-use,protected-access,unused-argument
import unittest

from mock import MagicMock, PropertyMock, patch

import botocore.exceptions
from botocore.stub import ANY, Stubber

from runway.cfngin.actions.base import BaseAction
from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.plan import Graph, Plan, Step
from runway.cfngin.providers.aws.default import Provider
from runway.cfngin.session_cache import get_session

from ..factories import MockProviderBuilder, mock_context

MOCK_VERSION = "01234abcdef"


class MockBlueprint(Blueprint):
    """Test blueprint."""

    VARIABLES = {
        "Param1": {"default": "default", "type": str},
    }

    @property
    def version(self):
        """Return version."""
        return MOCK_VERSION

    def create_template(self):
        """Create template."""


class TestBaseAction(unittest.TestCase):
    """Tests for runway.cfngin.actions.base.BaseAction."""

    def setUp(self):
        """Run before tests."""
        self.region = 'us-east-1'
        self.session = get_session(self.region)
        self.provider = Provider(self.session)

        self.config_no_persist = {
            'stacks': [
                {'name': 'stack1'},
                {'name': 'stack2',
                 'requires': ['stack1']}
            ]
        }

        self.config_persist = {
            'persistent_graph_key': 'test.json',
            'stacks': [
                {'name': 'stack1'},
                {'name': 'stack2',
                 'requires': ['stack1']}
            ]
        }

    def test_ensure_cfn_bucket_exists(self):
        """Test ensure cfn bucket exists."""
        session = get_session("us-east-1")
        provider = Provider(session)
        action = BaseAction(
            context=mock_context("mynamespace"),
            provider_builder=MockProviderBuilder(provider)
        )
        stubber = Stubber(action.s3_conn)
        stubber.add_response(
            "head_bucket",
            service_response={},
            expected_params={
                "Bucket": ANY,
            }
        )
        with stubber:
            action.ensure_cfn_bucket()

    def test_ensure_cfn_bucket_does_not_exist_us_east(self):
        """Test ensure cfn bucket does not exist us east."""
        session = get_session("us-east-1")
        provider = Provider(session)
        action = BaseAction(
            context=mock_context("mynamespace"),
            provider_builder=MockProviderBuilder(provider)
        )
        stubber = Stubber(action.s3_conn)
        stubber.add_client_error(
            "head_bucket",
            service_error_code="NoSuchBucket",
            service_message="Not Found",
            http_status_code=404,
        )
        stubber.add_response(
            "create_bucket",
            service_response={},
            expected_params={
                "Bucket": ANY,
            }
        )
        with stubber:
            action.ensure_cfn_bucket()

    def test_ensure_cfn_bucket_does_not_exist_us_west(self):
        """Test ensure cfn bucket does not exist us west."""
        session = get_session("us-west-1")
        provider = Provider(session)
        action = BaseAction(
            context=mock_context("mynamespace"),
            provider_builder=MockProviderBuilder(provider, region="us-west-1")
        )
        stubber = Stubber(action.s3_conn)
        stubber.add_client_error(
            "head_bucket",
            service_error_code="NoSuchBucket",
            service_message="Not Found",
            http_status_code=404,
        )
        stubber.add_response(
            "create_bucket",
            service_response={},
            expected_params={
                "Bucket": ANY,
                "CreateBucketConfiguration": {
                    "LocationConstraint": "us-west-1",
                }
            }
        )
        with stubber:
            action.ensure_cfn_bucket()

    def test_ensure_cfn_forbidden(self):
        """Test ensure cfn forbidden."""
        session = get_session("us-west-1")
        provider = Provider(session)
        action = BaseAction(
            context=mock_context("mynamespace"),
            provider_builder=MockProviderBuilder(provider)
        )
        stubber = Stubber(action.s3_conn)
        stubber.add_client_error(
            "head_bucket",
            service_error_code="AccessDenied",
            service_message="Forbidden",
            http_status_code=403,
        )
        with stubber:
            with self.assertRaises(botocore.exceptions.ClientError):
                action.ensure_cfn_bucket()

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    @patch('runway.cfngin.actions.base.BaseAction._stack_action',
           new_callable=PropertyMock)
    def test_generate_plan_no_persist_exclude(self, mock_stack_action,
                                              mock_tags):
        """Test generate plan no persist exclude."""
        mock_stack_action.return_value = MagicMock()
        mock_tags.return_value = {}
        context = mock_context(namespace='test',
                               extra_config_args=self.config_no_persist,
                               region=self.region)
        action = BaseAction(context=context,
                            provider_builder=MockProviderBuilder(
                                self.provider, region=self.region))

        plan = action._generate_plan(include_persistent_graph=False)

        mock_tags.assert_not_called()
        self.assertIsInstance(plan, Plan)
        # order is different between python2/3 so can't compare dicts
        result_graph_dict = plan.graph.to_dict()
        self.assertEqual(2, len(result_graph_dict))
        self.assertEqual(set(), result_graph_dict['stack1'])
        self.assertEqual(set(['stack1']), result_graph_dict['stack2'])
        self.assertEqual(BaseAction.DESCRIPTION, plan.description)
        self.assertTrue(plan.require_unlocked)

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    @patch('runway.cfngin.actions.base.BaseAction._stack_action',
           new_callable=PropertyMock)
    def test_generate_plan_no_persist_include(self, mock_stack_action,
                                              mock_tags):
        """Test generate plan no persist include."""
        mock_stack_action.return_value = MagicMock()
        mock_tags.return_value = {}
        context = mock_context(namespace='test',
                               extra_config_args=self.config_no_persist,
                               region=self.region)
        action = BaseAction(context=context,
                            provider_builder=MockProviderBuilder(
                                self.provider, region=self.region))

        plan = action._generate_plan(include_persistent_graph=True)

        mock_tags.assert_not_called()
        self.assertIsInstance(plan, Plan)
        # order is different between python2/3 so can't compare dicts
        result_graph_dict = plan.graph.to_dict()
        self.assertEqual(2, len(result_graph_dict))
        self.assertEqual(set(), result_graph_dict['stack1'])
        self.assertEqual(set(['stack1']), result_graph_dict['stack2'])
        self.assertEqual(BaseAction.DESCRIPTION, plan.description)
        self.assertTrue(plan.require_unlocked)

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    @patch('runway.cfngin.actions.base.BaseAction._stack_action',
           new_callable=PropertyMock)
    def test_generate_plan_with_persist_exclude(self, mock_stack_action,
                                                mock_tags):
        """Test generate plan with persist exclude."""
        mock_stack_action.return_value = MagicMock()
        mock_tags.return_value = {}
        context = mock_context(namespace='test',
                               extra_config_args=self.config_persist,
                               region=self.region)
        persist_step = Step.from_stack_name('removed', context)
        context._persistent_graph = Graph.from_steps([persist_step])
        action = BaseAction(context=context,
                            provider_builder=MockProviderBuilder(
                                self.provider, region=self.region))

        plan = action._generate_plan(include_persistent_graph=False)

        self.assertIsInstance(plan, Plan)
        # order is different between python2/3 so can't compare dicts
        result_graph_dict = plan.graph.to_dict()
        self.assertEqual(2, len(result_graph_dict))
        self.assertEqual(set(), result_graph_dict['stack1'])
        self.assertEqual(set(['stack1']), result_graph_dict['stack2'])
        self.assertEqual(BaseAction.DESCRIPTION, plan.description)
        self.assertTrue(plan.require_unlocked)

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    @patch('runway.cfngin.actions.base.BaseAction._stack_action',
           new_callable=PropertyMock)
    def test_generate_plan_with_persist_include(self, mock_stack_action,
                                                mock_tags):
        """Test generate plan with persist include."""
        mock_stack_action.return_value = MagicMock()
        mock_tags.return_value = {}
        context = mock_context(namespace='test',
                               extra_config_args=self.config_persist,
                               region=self.region)
        persist_step = Step.from_stack_name('removed', context)
        context._persistent_graph = Graph.from_steps([persist_step])
        action = BaseAction(context=context,
                            provider_builder=MockProviderBuilder(
                                self.provider, region=self.region))

        plan = action._generate_plan(include_persistent_graph=True)

        self.assertIsInstance(plan, Plan)
        mock_tags.assert_called_once()
        # order is different between python2/3 so can't compare dicts
        result_graph_dict = plan.graph.to_dict()
        self.assertEqual(3, len(result_graph_dict))
        self.assertEqual(set(), result_graph_dict['stack1'])
        self.assertEqual(set(['stack1']), result_graph_dict['stack2'])
        self.assertEqual(set(), result_graph_dict['removed'])
        self.assertEqual(BaseAction.DESCRIPTION, plan.description)
        self.assertTrue(plan.require_unlocked)

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    @patch('runway.cfngin.actions.base.BaseAction._stack_action',
           new_callable=PropertyMock)
    def test_generate_plan_with_persist_no_lock_req(self, mock_stack_action,
                                                    mock_tags):
        """Test generate plan with persist no lock req."""
        mock_stack_action.return_value = MagicMock()
        mock_tags.return_value = {}
        context = mock_context(namespace='test',
                               extra_config_args=self.config_persist,
                               region=self.region)
        persist_step = Step.from_stack_name('removed', context)
        context._persistent_graph = Graph.from_steps([persist_step])
        action = BaseAction(context=context,
                            provider_builder=MockProviderBuilder(
                                self.provider, region=self.region))

        plan = action._generate_plan(include_persistent_graph=True,
                                     require_unlocked=False)

        self.assertIsInstance(plan, Plan)
        mock_tags.assert_called_once()
        # order is different between python2/3 so can't compare dicts
        result_graph_dict = plan.graph.to_dict()
        self.assertEqual(3, len(result_graph_dict))
        self.assertEqual(set(), result_graph_dict['stack1'])
        self.assertEqual(set(['stack1']), result_graph_dict['stack2'])
        self.assertEqual(set(), result_graph_dict['removed'])
        self.assertEqual(BaseAction.DESCRIPTION, plan.description)
        self.assertFalse(plan.require_unlocked)

    def test_stack_template_url(self):
        """Test stack template url."""
        context = mock_context("mynamespace")
        blueprint = MockBlueprint(name="myblueprint", context=context)

        region = "us-east-1"
        endpoint = "https://example.com"
        session = get_session(region)
        provider = Provider(session)
        action = BaseAction(
            context=context,
            provider_builder=MockProviderBuilder(provider, region=region)
        )

        with patch('runway.cfngin.actions.base.get_s3_endpoint',
                   autospec=True, return_value=endpoint):
            self.assertEqual(
                action.stack_template_url(blueprint),
                "%s/%s/stack_templates/%s/%s-%s.json" % (
                    endpoint,
                    "stacker-mynamespace",
                    "mynamespace-myblueprint",
                    "myblueprint",
                    MOCK_VERSION
                )
            )



"""Test that multiple CDK stacks does not prompt a failure"""
from runway.util import change_dir

from integration_tests.test_cdk.test_cdk import CDK
from integration_tests.util import run_command


class TestMultipleStacks(CDK):
    """Test deploying multiple stacks and ensure all are deployed"""

    TEST_NAME = __name__
    module_dir = 'multiple-stacks-app.cdk'

    def deploy(self):
        """Deploy provider."""
        self.copy_fixture(self.module_dir)
        self.copy_runway('multiple-stacks')
        with change_dir(self.cdk_test_dir):
            return run_command(['runway', 'deploy'])

    def run(self):
        """Run tests."""
        self.clean()
        self.set_env_var('CI', '1')
        assert self.deploy() == 0, '{}: Multiple Stacks failed'.format(__name__)

    def teardown(self):
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        self.delete_venv(self.module_dir)
        with change_dir(self.cdk_test_dir):
            run_command(['runway', 'destroy'])
        self.clean()
        self.unset_env_var('CI')



"""Test deploying a base line static site."""
import os

from runway.util import change_dir

from integration_tests.test_staticsite.test_staticsite import StaticSite
from integration_tests.util import run_command


class TestBasicSite(StaticSite):
    """Test deploying a base line static site."""

    TEST_NAME = __name__
    module_dir = 'basic-site'

    def deploy(self):
        """Deploy provider."""
        os.mkdir(self.staticsite_test_dir)
        os.mkdir(os.path.join(self.staticsite_test_dir, self.module_dir))
        os.mkdir(os.path.join(self.staticsite_test_dir, self.module_dir, 'build'))
        self.copy_runway('basic-site')
        with change_dir(self.staticsite_test_dir):
            return run_command(['runway', 'deploy'])

    def run(self):
        """Run tests."""
        self.clean()
        self.set_env_var('CI', '1')
        assert self.deploy() == 0, '{}: Basic Site failed'.format(__name__)

    def teardown(self):
        self.logger.info('Tearing down: %s', self.TEST_NAME)
        self.delete_venv(self.module_dir)
        with change_dir(self.staticsite_test_dir):
            run_command(['runway', 'destroy'])
        self.clean()



"""Tests runway/commands/modules_command.py."""
import os
import unittest
from copy import deepcopy
from os import path

import yaml
from mock import patch
from moto import mock_sts

from runway.commands.modules_command import (select_modules_to_run,
                                             validate_environment)


def module_tag_config():
    """Return a runway.yml file for testing module tags."""
    fixture_dir = path.join(
        path.dirname(path.dirname(path.realpath(__file__))),
        'fixtures'
    )
    with open(path.join(fixture_dir, 'tag.runway.yml'), 'r') as stream:
        return yaml.safe_load(stream)


class ModulesCommandTestCase(unittest.TestCase):
    """Test runway/commands/modules_command.py."""

    tag_yml = module_tag_config()

    def test_select_modules_to_run_tag_test_app(self):
        """tag=[app:test-app] should return 2 modules."""
        config = deepcopy(ModulesCommandTestCase.tag_yml)
        tags = ['app:test-app']

        result = [
            select_modules_to_run(deployment, tags)
            for deployment in config['deployments']
        ]
        self.assertEqual(len(result[0]['modules']), 1)
        self.assertEqual(result[0]['modules'][0]['path'], 'sampleapp1.cfn')
        self.assertEqual(len(result[1]['modules']), 0)
        self.assertEqual(len(result[2]['modules']), 1)
        self.assertEqual(result[2]['modules'][0]['path'], 'sampleapp4.cfn')

    def test_select_modules_to_run_tag_iac(self):
        """tag=[tier:iac] should return 2 modules."""
        config = deepcopy(ModulesCommandTestCase.tag_yml)
        tags = ['tier:iac']

        result = [
            select_modules_to_run(deployment, tags)
            for deployment in config['deployments']
        ]
        self.assertEqual(len(result[0]['modules']), 2)
        self.assertEqual(result[0]['modules'][0]['path'], 'sampleapp1.cfn')
        self.assertEqual(result[0]['modules'][1]['path'], 'sampleapp2.cfn')
        self.assertEqual(len(result[1]['modules']), 0)
        self.assertEqual(len(result[2]['modules']), 0)

    def test_select_modules_to_run_two_tags(self):
        """tag=[tier:iac, app:test-app] should return 1 module."""
        config = deepcopy(ModulesCommandTestCase.tag_yml)
        tags = ['tier:iac', 'app:test-app']
        result = [
            select_modules_to_run(deployment, tags)
            for deployment in config['deployments']
        ]
        self.assertEqual(len(result[0]['modules']), 1)
        self.assertEqual(result[0]['modules'][0]['path'], 'sampleapp1.cfn')
        self.assertEqual(len(result[1]['modules']), 0)
        self.assertEqual(len(result[2]['modules']), 0)

    def test_select_modules_to_run_no_tags(self):
        """tag=[] should request input."""
        config = deepcopy(ModulesCommandTestCase.tag_yml)
        user_input = ['1']
        with patch('runway.commands.modules_command.input',
                   side_effect=user_input):
            result = select_modules_to_run(
                config['deployments'][0], []
            )
        self.assertEqual(result['modules'][0],
                         config['deployments'][0]['modules'][0])

    def test_select_modules_to_run_no_tags_ci(self):
        """tag=[], ci=true should not request input and return everything."""
        config = deepcopy(ModulesCommandTestCase.tag_yml)
        result = [
            select_modules_to_run(deployment, [], ci='true')
            for deployment in config['deployments']
        ]
        self.assertEqual(result, config['deployments'])

    def test_select_modules_to_run_destroy(self):
        """command=destroy should only prompt with no tag of ci if one module."""
        config = deepcopy(ModulesCommandTestCase.tag_yml)
        user_input = ['y', '1']
        with patch('runway.commands.modules_command.input',
                   side_effect=user_input):
            result_single_no_tag = select_modules_to_run(
                deepcopy(config['deployments'][1]), [], command='destroy'
            )
            result_no_tag = select_modules_to_run(
                deepcopy(config['deployments'][0]), [], command='destroy'
            )
        self.assertEqual(result_single_no_tag['modules'][0],
                         config['deployments'][1]['modules'][0])
        self.assertEqual(result_no_tag['modules'][0],
                         config['deployments'][0]['modules'][0])
        result_tag = select_modules_to_run(
            deepcopy(config['deployments'][0]), ['app:test-app'],
            command='destroy'
        )
        self.assertEqual(result_tag['modules'][0],
                         config['deployments'][0]['modules'][0])
        result_tag_ci = select_modules_to_run(
            deepcopy(config['deployments'][0]), [], command='destroy',
            ci='true'
        )
        self.assertEqual(result_tag_ci['modules'],
                         config['deployments'][0]['modules'])


class TestValidateEnvironment(object):
    """Tests for validate_environment."""

    MOCK_ACCOUNT_ID = '123456789012'

    def test_bool_match(self):
        """True bool should match."""
        assert validate_environment('test_module', True, os.environ)

    def test_bool_not_match(self):
        """False bool should not match."""
        assert not validate_environment('test_module', False, os.environ)

    @mock_sts
    def test_list_match(self):
        """Env in list should match."""
        assert validate_environment('test_module',
                                    [self.MOCK_ACCOUNT_ID + '/us-east-1'],
                                    os.environ)

    @mock_sts
    def test_list_not_match(self):
        """Env not in list should not match."""
        assert not validate_environment('test_module',
                                        [self.MOCK_ACCOUNT_ID + '/us-east-2'],
                                        os.environ)

    @mock_sts
    def test_str_match(self):
        """Env in string should match."""
        assert validate_environment('test_module',
                                    self.MOCK_ACCOUNT_ID + '/us-east-1',
                                    os.environ)

    @mock_sts
    def test_str_not_match(self):
        """Env not in string should not match."""
        assert not validate_environment('test_module',
                                        self.MOCK_ACCOUNT_ID + '/us-east-2',
                                        os.environ)



"""Gets information on the CloudFormation stacks based on the given config."""
from ...actions import info
from .base import BaseCommand


class Info(BaseCommand):
    """Info subcommand."""

    name = "info"
    description = __doc__

    def add_arguments(self, parser):
        """Add arguments."""
        super(Info, self).add_arguments(parser)
        self._add_argument_stacks(parser)

    def run(self, options):
        """Run the command."""
        super(Info, self).run(options)
        action = info.Action(options.context,
                             provider_builder=options.provider_builder)

        action.execute()

    def get_context_kwargs(self, options):
        """Return a dictionary of kwargs that will be used with the Context.

        This allows commands to pass in any specific arguments they define to
        the context.

        Args:
            options (:class:`argparse.Namespace`): arguments that have been
                passed via the command line

        Returns:
            Dict[str, Any]: Dictionary that will be passed to Context
            initializer as kwargs.

        """
        return {"stack_names": options.stacks}



"""Tests for runway.cfngin.lookups.handlers.kms."""
import codecs
import unittest

import boto3
from botocore.stub import Stubber
from mock import patch

from runway.cfngin.lookups.handlers.kms import KmsLookup

from ...factories import SessionStub, mock_provider

REGION = 'us-east-1'


class TestKMSHandler(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.kms.KmsLookup."""

    client = boto3.client('kms', region_name=REGION,
                          # bypass the need to have these in the env
                          aws_access_key_id='testing',
                          aws_secret_access_key='testing')

    def setUp(self):
        """Run before tests."""
        self.stubber = Stubber(self.client)
        self.provider = mock_provider(region=REGION)
        self.secret = b'my secret'

    @patch("runway.cfngin.lookups.handlers.kms.get_session",
           return_value=SessionStub(client))
    def test_kms_handler(self, _mock_client):
        """Test kms handler."""
        self.stubber.add_response('decrypt', {'Plaintext': self.secret},
                                  {'CiphertextBlob': codecs.decode(self.secret,
                                                                   'base64')})

        with self.stubber:
            self.assertEqual(self.secret,
                             KmsLookup.handle(value=self.secret.decode(),
                                              provider=self.provider))
            self.stubber.assert_no_pending_responses()

    @patch("runway.cfngin.lookups.handlers.kms.get_session",
           return_value=SessionStub(client))
    def test_kms_handler_with_region(self, _mock_client):
        """Test kms handler with region."""
        value = '{}@{}'.format(REGION, self.secret.decode())

        self.stubber.add_response('decrypt', {'Plaintext': self.secret},
                                  {'CiphertextBlob': codecs.decode(self.secret,
                                                                   'base64')})

        with self.stubber:
            self.assertEqual(self.secret,
                             KmsLookup.handle(value=value,
                                              provider=self.provider))
            self.stubber.assert_no_pending_responses()



"""CFNgin target."""


class Target(object):  # pylint: disable=too-few-public-methods
    """A "target" is just a node in the graph that only specify dependencies.

    These can be useful as a means of logically grouping a set of stacks
    together that can be targeted with the ``targets`` option.

    Attributes:
        logging (bool): Whether logging is enabled.
        name (str): Name of the target (stack) taken from the definition.
        required_by (List[str]): List of target (stack) names that depend on
            this stack.
        requires (List[str]): List of target (stack) names this target (stack)
            depends on.

    """

    def __init__(self, definition):
        """Instantiate class.

        Args:
            definition (:class:`runway.cfngin.config.Stack`): Stack definition
                for the target.

        """
        self.name = definition.name
        self.requires = definition.requires or []
        self.required_by = definition.required_by or []
        self.logging = False



"""Tests for runway.cfngin.lookups.handlers.ssmstore."""
import unittest

import boto3
import mock
from botocore.stub import Stubber
from six import string_types

from runway.cfngin.lookups.handlers.ssmstore import SsmstoreLookup

from ...factories import SessionStub


class TestSSMStoreHandler(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.ssmstore.SsmstoreLookup."""

    client = boto3.client('ssm', region_name='us-east-1',
                          # bypass the need to have these in the env
                          aws_access_key_id='testing',
                          aws_secret_access_key='testing')

    def setUp(self):
        """Run before tests."""
        self.stubber = Stubber(self.client)
        self.get_parameters_response = {
            'Parameters': [
                {
                    'Name': 'ssmkey',
                    'Type': 'String',
                    'Value': 'ssmvalue'
                }
            ],
            'InvalidParameters': [
                'invalid_ssm_param'
            ]
        }
        self.invalid_get_parameters_response = {
            'InvalidParameters': [
                'ssmkey'
            ]
        }
        self.expected_params = {
            'Names': ['ssmkey'],
            'WithDecryption': True
        }
        self.ssmkey = "ssmkey"
        self.ssmvalue = "ssmvalue"

    @mock.patch('runway.cfngin.lookups.handlers.ssmstore.get_session',
                return_value=SessionStub(client))
    def test_ssmstore_handler(self, _mock_client):
        """Test ssmstore handler."""
        self.stubber.add_response('get_parameters',
                                  self.get_parameters_response,
                                  self.expected_params)
        with self.stubber:
            value = SsmstoreLookup.handle(self.ssmkey)
            self.assertEqual(value, self.ssmvalue)
            self.assertIsInstance(value, string_types)

    @mock.patch('runway.cfngin.lookups.handlers.ssmstore.get_session',
                return_value=SessionStub(client))
    def test_ssmstore_invalid_value_handler(self, _mock_client):
        """Test ssmstore invalid value handler."""
        self.stubber.add_response('get_parameters',
                                  self.invalid_get_parameters_response,
                                  self.expected_params)
        with self.stubber:
            try:
                SsmstoreLookup.handle(self.ssmkey)
            except ValueError:
                assert True

    @mock.patch('runway.cfngin.lookups.handlers.ssmstore.get_session',
                return_value=SessionStub(client))
    def test_ssmstore_handler_with_region(self, _mock_client):
        """Test ssmstore handler with region."""
        self.stubber.add_response('get_parameters',
                                  self.get_parameters_response,
                                  self.expected_params)
        region = "us-east-1"
        temp_value = "%s@%s" % (region, self.ssmkey)
        with self.stubber:
            value = SsmstoreLookup.handle(temp_value)
            self.assertEqual(value, self.ssmvalue)



"""Hook data lookup."""
# pylint: disable=arguments-differ,unused-argument
from runway.lookups.handlers.base import LookupHandler

TYPE_NAME = "hook_data"


class HookDataLookup(LookupHandler):
    """Hook data lookup."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Return the value of a key for a given hook in hook_data.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        Format of value:

            <hook_name>::<key>

        """
        try:
            hook_name, key = value.split("::")
        except ValueError:
            raise ValueError("Invalid value for hook_data: %s. Must be in "
                             "<hook_name>::<key> format." % value)

        return context.hook_data[hook_name][key]



"""Tests for Serverless."""
from __future__ import print_function
import os
from send2trash import send2trash

from integration_tests.integration_test import IntegrationTest
from integration_tests.util import (import_tests, execute_tests)


class Serverless(IntegrationTest):
    """Base class for all Serverless testing."""

    base_dir = os.path.abspath(os.path.dirname(__file__))
    templates_dir = os.path.join(base_dir, 'templates')
    tests_dir = os.path.join(base_dir, 'tests')

    def clean(self):
        """Cleanup serverless folder."""
        # cleanup any leftover .serverless folder
        serverless_folder = os.path.join(self.templates_dir,
                                         self.template_name,
                                         '.serverless')
        if os.path.isdir(serverless_folder):
            self.logger.debug('send2trash: "%s"', serverless_folder)
            send2trash(serverless_folder)

    def run(self):
        """Find all Serverless tests and run them."""
        self.set_env_var('CI', '1')
        import_tests(self.logger, self.base_dir, 'serverless_test')
        serverless_test = Serverless.__subclasses__()[0]

        tests = []
        for template in os.listdir(self.templates_dir):
            if os.path.isdir(os.path.join(self.templates_dir, template)):
                self.logger.info('Found template "%s"', template)
                test = serverless_test(template,
                                       self.templates_dir,
                                       self.environment,
                                       self.logger)
                tests.append(test)
            else:
                self.logger.warning('"%s" is not a directory, skipping...', template)

        if not tests:
            raise Exception('No tests were found.')

        err_count = execute_tests(tests, self.logger)
        assert err_count == 0  # assert that all subtests were successful
        return err_count

    def teardown(self):
        """Teardown resources create during init."""
        self.unset_env_var('CI')
        # all resources should have been torn down in ServerlessTest class



"""Pytest fixtures and plugins."""
import os

import py
import pytest

from runway.cfngin.dag import DAG


@pytest.fixture(scope="package")
def cfngin_fixture_dir():
    """Path to the fixture directory."""
    path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        'fixtures')
    return py.path.local(path)  # pylint: disable=no-member


@pytest.fixture
def empty_dag():
    """Create an empty DAG."""
    return DAG()


@pytest.fixture
def basic_dag():
    """Create a basic DAG."""
    dag = DAG()
    dag.from_dict({'a': ['b', 'c'],
                   'b': ['d'],
                   'c': ['d'],
                   'd': []})
    return dag



#!/usr/bin/env python
"""Module with static website bucket and CloudFront distribution."""
from __future__ import print_function

import hashlib
# https://github.com/PyCQA/pylint/issues/73
from distutils.version import LooseVersion  # noqa pylint: disable=no-name-in-module,import-error
from past.builtins import basestring

import awacs.s3
import awacs.sts
from awacs.aws import Action, Allow, Policy, PolicyDocument, Principal, Statement

import troposphere
from troposphere import (
    AWSProperty, And, Equals, If, Join, Not, NoValue, Output, Select,
    awslambda, cloudfront, iam, s3
)

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.blueprints.variables.types import CFNCommaDelimitedList, CFNString
from runway.cfngin.context import Context

IAM_ARN_PREFIX = 'arn:aws:iam::aws:policy/service-role/'
if LooseVersion(troposphere.__version__) == LooseVersion('2.4.0'):
    # pylint: disable=ungrouped-imports
    from troposphere.validators import boolean, priceclass_type

    class S3OriginConfig(AWSProperty):
        """Backported s3 origin config class for broken troposphere release."""

        props = {
            'OriginAccessIdentity': (basestring, False),
        }

    class Origin(AWSProperty):
        """Backported origin config class for broken troposphere release."""

        props = {
            'CustomOriginConfig': (cloudfront.CustomOriginConfig, False),
            'DomainName': (basestring, True),
            'Id': (basestring, True),
            'OriginCustomHeaders': ([cloudfront.OriginCustomHeader], False),
            'OriginPath': (basestring, False),
            'S3OriginConfig': (S3OriginConfig, False),
        }

    class DistributionConfig(AWSProperty):
        """Backported cf config class for broken troposphere release."""

        props = {
            'Aliases': (list, False),
            'CacheBehaviors': ([cloudfront.CacheBehavior], False),
            'Comment': (basestring, False),
            'CustomErrorResponses': ([cloudfront.CustomErrorResponse], False),
            'DefaultCacheBehavior': (cloudfront.DefaultCacheBehavior, True),
            'DefaultRootObject': (basestring, False),
            'Enabled': (boolean, True),
            'HttpVersion': (basestring, False),
            'IPV6Enabled': (boolean, False),
            'Logging': (cloudfront.Logging, False),
            'Origins': ([Origin], True),
            'PriceClass': (priceclass_type, False),
            'Restrictions': (cloudfront.Restrictions, False),
            'ViewerCertificate': (cloudfront.ViewerCertificate, False),
            'WebACLId': (basestring, False),
        }


class StaticSite(Blueprint):  # pylint: disable=too-few-public-methods
    """Stacker blueprint for creating S3 bucket and CloudFront distribution."""

    VARIABLES = {
        'AcmCertificateArn': {'type': CFNString,
                              'default': '',
                              'description': '(Optional) Cert ARN for site'},
        'Aliases': {'type': CFNCommaDelimitedList,
                    'default': '',
                    'description': '(Optional) Domain aliases the '
                                   'distribution'},
        'DisableCloudFront': {'type': CFNString,
                              'default': '',
                              'description': 'Whether to disable CF'},
        'LogBucketName': {'type': CFNString,
                          'default': '',
                          'description': 'S3 bucket for CF logs'},
        'PriceClass': {'type': CFNString,
                       'default': 'PriceClass_100',  # US/Europe
                       'description': 'CF price class for the distribution.'},
        'RewriteDirectoryIndex': {'type': CFNString,
                                  'default': '',
                                  'description': '(Optional) File name to '
                                                 'append to directory '
                                                 'requests.'},
        'WAFWebACL': {'type': CFNString,
                      'default': '',
                      'description': '(Optional) WAF id to associate with the '
                                     'distribution.'},
        'custom_error_responses': {'type': list,
                                   'default': [],
                                   'description': '(Optional) Custom error '
                                                  'responses.'},
        'lambda_function_associations': {'type': list,
                                         'default': [],
                                         'description': '(Optional) Lambda '
                                                        'function '
                                                        'associations.'},
    }

    def create_template(self):
        """Create template (main function called by Stacker)."""
        self.template.set_version('2010-09-09')
        self.template.set_description('Static Website - Bucket and Distribution')

        self.add_template_conditions()

        # Resources
        bucket = self.add_bucket()
        bucket_policy = self.add_bucket_policy(bucket) # noqa pylint: disable=unused-variable
        oai = self.add_origin_access_identity()
        allow_access = self.allow_cloudfront_access_on_bucket(bucket, oai)
        rewrite_role = self.add_index_rewrite_role()
        index_rewrite = self.add_cloudfront_directory_index_rewrite(rewrite_role)
        index_rewrite_version = self.add_cloudfront_directory_index_rewrite_version(
            index_rewrite
        )
        lambda_function_associations = self.get_lambda_associations(index_rewrite_version)
        distribution_options = self.get_cloudfront_distribution_options(
            bucket,
            oai,
            lambda_function_associations
        )
        distribution = self.add_cloudfront_distribution( # noqa pylint: disable=unused-variable
            allow_access,
            distribution_options
        )

    def add_template_conditions(self):
        """Add Template Conditions."""
        variables = self.get_variables()

        self.template.add_condition(
            'AcmCertSpecified',
            And(Not(Equals(variables['AcmCertificateArn'].ref, '')),
                Not(Equals(variables['AcmCertificateArn'].ref, 'undefined')))
        )
        self.template.add_condition(
            'AliasesSpecified',
            And(Not(Equals(Select(0, variables['Aliases'].ref), '')),
                Not(Equals(Select(0, variables['Aliases'].ref), 'undefined')))
        )
        self.template.add_condition(
            'CFEnabled',
            Not(Equals(variables['DisableCloudFront'].ref, 'true'))
        )
        self.template.add_condition(
            'CFDisabled',
            Equals(variables['DisableCloudFront'].ref, 'true')
        )
        self.template.add_condition(
            'CFLoggingEnabled',
            And(Not(Equals(variables['LogBucketName'].ref, '')),
                Not(Equals(variables['LogBucketName'].ref, 'undefined')))
        )
        self.template.add_condition(
            'DirectoryIndexSpecified',
            And(Not(Equals(variables['RewriteDirectoryIndex'].ref, '')),
                Not(Equals(variables['RewriteDirectoryIndex'].ref, 'undefined')))  # noqa
        )
        self.template.add_condition(
            'CFEnabledAndDirectoryIndexSpecified',
            And(Not(Equals(variables['RewriteDirectoryIndex'].ref, '')),
                Not(Equals(variables['RewriteDirectoryIndex'].ref, 'undefined')), # noqa
                Not(Equals(variables['DisableCloudFront'].ref, 'true')))
        )
        self.template.add_condition(
            'WAFNameSpecified',
            And(Not(Equals(variables['WAFWebACL'].ref, '')),
                Not(Equals(variables['WAFWebACL'].ref, 'undefined')))
        )

    def get_lambda_associations(self, directory_index_rewrite_version):
        """Retrieve any lambda associations from the instance variables.

        Keyword Args:
            directory_index_rewrite_version (dict): The directory index rewrite lambda version
                resource

        Return:
            array: Array of lambda function association variables
        """
        variables = self.get_variables()

        # If custom associations defined, use them
        if variables['lambda_function_associations']:
            return [
                cloudfront.LambdaFunctionAssociation(
                    EventType=x['type'],
                    LambdaFunctionARN=x['arn']
                ) for x in variables['lambda_function_associations']
            ]

        # otherwise fallback to pure CFN condition
        return If(
            'DirectoryIndexSpecified',
            [cloudfront.LambdaFunctionAssociation(
                EventType='origin-request',
                LambdaFunctionARN=directory_index_rewrite_version.ref()
            )],
            NoValue
        )

    def get_cloudfront_distribution_options(self, bucket, oai, lambda_function_associations):
        """Retrieve the options for our CloudFront distribution.

        Keyword Args:
            bucket (dict): The bucket resource
            oai (dict): The origin access identity resource
            lambda_function_associations (array): The lambda function association array

        Return:
            dict: The CloudFront Distribution Options

        """
        variables = self.get_variables()
        return {
            'Aliases': If(
                'AliasesSpecified',
                variables['Aliases'].ref,
                NoValue
            ),
            'Origins': [
                get_cf_origin_class()(
                    DomainName=Join(
                        '.',
                        [bucket.ref(),
                         's3.amazonaws.com']),
                    S3OriginConfig=get_s3_origin_conf_class()(
                        OriginAccessIdentity=Join(
                            '',
                            ['origin-access-identity/cloudfront/',
                             oai.ref()])
                    ),
                    Id='S3Origin'
                )
            ],
            'DefaultCacheBehavior': cloudfront.DefaultCacheBehavior(
                AllowedMethods=['GET', 'HEAD'],
                Compress=False,
                DefaultTTL='86400',
                ForwardedValues=cloudfront.ForwardedValues(
                    Cookies=cloudfront.Cookies(Forward='none'),
                    QueryString=False,
                ),
                LambdaFunctionAssociations=lambda_function_associations,
                TargetOriginId='S3Origin',
                ViewerProtocolPolicy='redirect-to-https'
            ),
            'DefaultRootObject': 'index.html',
            'Logging': If(
                'CFLoggingEnabled',
                cloudfront.Logging(
                    Bucket=Join('.',
                                [variables['LogBucketName'].ref,
                                 's3.amazonaws.com'])
                ),
                NoValue
            ),
            'PriceClass': variables['PriceClass'].ref,
            'Enabled': True,
            'WebACLId': If(
                'WAFNameSpecified',
                variables['WAFWebACL'].ref,
                NoValue
            ),
            'ViewerCertificate': If(
                'AcmCertSpecified',
                cloudfront.ViewerCertificate(
                    AcmCertificateArn=variables['AcmCertificateArn'].ref,
                    SslSupportMethod='sni-only'
                ),
                NoValue
            )
        }

    def add_origin_access_identity(self):
        """Add the origin access identity resource to the template.

        Returns:
            dict: The OAI resource

        """
        return self.template.add_resource(
            cloudfront.CloudFrontOriginAccessIdentity(
                'OAI',
                Condition='CFEnabled',
                CloudFrontOriginAccessIdentityConfig=cloudfront.CloudFrontOriginAccessIdentityConfig(  # noqa pylint: disable=line-too-long
                    Comment='CF access to website'
                )
            )
        )

    def add_bucket_policy(self, bucket):
        """Add a policy to the bucket if CloudFront is disabled. Ensure PublicRead.

        Keyword Args:
            bucket (dict): The bucket resource to place the policy

        Returns:
            dict: The Bucket Policy Resource

        """
        return self.template.add_resource(
            s3.BucketPolicy(
                'BucketPolicy',
                Bucket=bucket.ref(),
                Condition='CFDisabled',
                PolicyDocument=Policy(
                    Version="2012-10-17",
                    Statement=[
                        Statement(
                            Effect=Allow,
                            Principal=Principal('*'),
                            Action=[Action('s3', 'getObject')],
                            Resource=[
                                Join('', [bucket.get_att('Arn'), '/*'])
                            ],
                        )
                    ]
                )
            )
        )

    def add_bucket(self):
        """Add the bucket resource along with an output of it's name / website url.

        Returns:
            dict: The bucket resource

        """
        bucket = self.template.add_resource(
            s3.Bucket(
                'Bucket',
                AccessControl=If('CFEnabled', s3.Private, s3.PublicRead),
                LifecycleConfiguration=s3.LifecycleConfiguration(
                    Rules=[
                        s3.LifecycleRule(
                            NoncurrentVersionExpirationInDays=90,
                            Status='Enabled'
                        )
                    ]
                ),
                VersioningConfiguration=s3.VersioningConfiguration(
                    Status='Enabled'
                ),
                WebsiteConfiguration=s3.WebsiteConfiguration(
                    IndexDocument='index.html',
                    ErrorDocument='error.html'
                )
            )
        )
        self.template.add_output(Output(
            'BucketName',
            Description='Name of website bucket',
            Value=bucket.ref()
        ))
        self.template.add_output(Output(
            'BucketWebsiteURL',
            Condition="CFDisabled",
            Description='URL of the bucket website',
            Value=bucket.get_att('WebsiteURL')
        ))
        return bucket

    def allow_cloudfront_access_on_bucket(self, bucket, oai):
        """Given a bucket and oai resource add cloudfront access to the bucket.

        Keyword Args:
            bucket (dict): A bucket resource
            oai (dict): An Origin Access Identity resource

        Return:
            dict: The CloudFront Bucket access resource
        """
        return self.template.add_resource(
            s3.BucketPolicy(
                'AllowCFAccess',
                Bucket=bucket.ref(),
                Condition='CFEnabled',
                PolicyDocument=PolicyDocument(
                    Version='2012-10-17',
                    Statement=[
                        Statement(
                            Action=[awacs.s3.GetObject],
                            Effect=Allow,
                            Principal=Principal(
                                'CanonicalUser',
                                oai.get_att('S3CanonicalUserId')
                            ),
                            Resource=[
                                Join('', [bucket.get_att('Arn'),
                                          '/*'])
                            ]
                        )
                    ]
                )
            )
        )

    def add_index_rewrite_role(self):
        """Add an index rewrite role to the template.

        Return:
            dict: The index rewrite role
        """
        return self.template.add_resource(
            iam.Role(
                'CFDirectoryIndexRewriteRole',
                Condition='CFEnabledAndDirectoryIndexSpecified',
                AssumeRolePolicyDocument=PolicyDocument(
                    Version='2012-10-17',
                    Statement=[
                        Statement(
                            Effect=Allow,
                            Action=[awacs.sts.AssumeRole],
                            Principal=Principal('Service',
                                                ['lambda.amazonaws.com',
                                                 'edgelambda.amazonaws.com'])
                        )
                    ]
                ),
                ManagedPolicyArns=[
                    IAM_ARN_PREFIX + 'AWSLambdaBasicExecutionRole'
                ]
            )
        )

    def add_cloudfront_directory_index_rewrite(self, role):
        """Add an index CloudFront directory index rewrite lambda function to the template.

        Keyword Args:
            role (dict): The index rewrite role resource

        Return:
            dict: The CloudFront directory index rewrite lambda function resource
        """
        variables = self.get_variables()
        return self.template.add_resource(
            awslambda.Function(
                'CFDirectoryIndexRewrite',
                Condition='CFEnabledAndDirectoryIndexSpecified',
                Code=awslambda.Code(
                    ZipFile=Join(
                        '',
                        ["'use strict';\n",
                         "exports.handler = async function(event, context) {\n",
                         "\n",
                         "    // Extract the request from the CloudFront event that is sent to Lambda@Edge\n",  # noqa pylint: disable=line-too-long
                         "    var request = event.Records[0].cf.request;\n",
                         "    // Extract the URI from the request\n",
                         "    var olduri = request.uri;\n",
                         "    // Match any '/' that occurs at the end of a URI. Replace it with a default index\n",  # noqa pylint: disable=line-too-long
                         "    var newuri = olduri.replace(/\\/$/, '\\/",
                         variables['RewriteDirectoryIndex'].ref,
                         "');\n",  # noqa
                         "    // Log the URI as received by CloudFront and the new URI to be used to fetch from origin\n",  # noqa pylint: disable=line-too-long
                         "    console.log(\"Old URI: \" + olduri);\n",
                         "    console.log(\"New URI: \" + newuri);\n",
                         "    // Replace the received URI with the URI that includes the index page\n",  # noqa pylint: disable=line-too-long
                         "    request.uri = newuri;\n",
                         "    // Return to CloudFront\n",
                         "    return request;\n",
                         "\n",
                         "};\n"]
                    )
                ),
                Description='Rewrites CF directory HTTP requests to default page',  # noqa
                Handler='index.handler',
                Role=role.get_att('Arn'),
                Runtime='nodejs10.x'
            )
        )

    def add_cloudfront_directory_index_rewrite_version(self, directory_index_rewrite):
        """Add a specific version to the directory index rewrite lambda.

        Keyword Args:
            directory_index_rewrite (dict): The directory index rewrite lambda resource

        Return:
            dict: The CloudFront directory index rewrite version

        """
        # Generating a unique resource name here for the Lambda version, so it
        # updates automatically if the lambda code changes
        code_hash = hashlib.md5(
            str(directory_index_rewrite.properties['Code'].properties['ZipFile'].to_dict()).encode()  # noqa pylint: disable=line-too-long
        ).hexdigest()

        return self.template.add_resource(
            awslambda.Version(
                'CFDirectoryIndexRewriteVer' + code_hash,
                Condition='CFEnabledAndDirectoryIndexSpecified',
                FunctionName=directory_index_rewrite.ref()
            )
        )

    def add_cloudfront_distribution(
            self,
            allow_cloudfront_access,
            cloudfront_distribution_options
    ):
        """Add the CloudFront distribution to the template / output the id and domain name.

        Keyword Args:
            allow_cloudfront_access (dict): Allow bucket access resource
            cloudfront_distribution_options (dict): The distribution options

        Return:
            dict: The CloudFront Distribution resource

        """
        distribution = self.template.add_resource(
            get_cf_distribution_class()(
                'CFDistribution',
                Condition='CFEnabled',
                DependsOn=allow_cloudfront_access.title,
                DistributionConfig=get_cf_distro_conf_class()(
                    **cloudfront_distribution_options
                )
            )
        )
        self.template.add_output(Output(
            'CFDistributionId',
            Condition='CFEnabled',
            Description='CloudFront distribution ID',
            Value=distribution.ref()
        ))
        self.template.add_output(
            Output(
                'CFDistributionDomainName',
                Condition='CFEnabled',
                Description='CloudFront distribution domain name',
                Value=distribution.get_att('DomainName')
            )
        )
        return distribution


def get_cf_distribution_class():
    """Return the correct troposphere CF distribution class."""
    if LooseVersion(troposphere.__version__) == LooseVersion('2.4.0'):
        cf_dist = cloudfront.Distribution
        cf_dist.props['DistributionConfig'] = (DistributionConfig, True)
        return cf_dist
    return cloudfront.Distribution


def get_cf_distro_conf_class():
    """Return the correct troposphere CF distribution class."""
    if LooseVersion(troposphere.__version__) == LooseVersion('2.4.0'):
        return DistributionConfig
    return cloudfront.DistributionConfig


def get_cf_origin_class():
    """Return the correct Origin class for troposphere."""
    if LooseVersion(troposphere.__version__) == LooseVersion('2.4.0'):
        return Origin
    return cloudfront.Origin


def get_s3_origin_conf_class():
    """Return the correct S3 Origin Config class for troposphere."""
    if LooseVersion(troposphere.__version__) > LooseVersion('2.4.0'):
        return cloudfront.S3OriginConfig
    if LooseVersion(troposphere.__version__) == LooseVersion('2.4.0'):
        return S3OriginConfig
    return cloudfront.S3Origin


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    print(StaticSite('test', Context({"namespace": "test"}), None).to_json())



"""Diffs the config against the currently running CloudFormation stacks.

Sometimes small changes can have big impacts.  Run "cfngin diff" before
"cfngin build" to detect bad things(tm) from happening in advance!

"""
from ...actions import diff
from .base import BaseCommand


class Diff(BaseCommand):
    """Diff subcommand."""

    name = "diff"
    description = __doc__

    def add_arguments(self, parser):
        """Add arguments."""
        super(Diff, self).add_arguments(parser)
        self._add_argument_stacks(parser)
        parser.add_argument("--force", action="append", default=[],
                            metavar="STACKNAME", type=str,
                            help="If a stackname is provided to --force, it "
                                 "will be diffed, even if it is locked in "
                                 "the config.")

    def run(self, options):
        """Run the command."""
        super(Diff, self).run(options)
        action = diff.Action(options.context,
                             provider_builder=options.provider_builder)
        action.execute()

    def get_context_kwargs(self, options):
        """Return a dictionary of kwargs that will be used with the Context.

        This allows commands to pass in any specific arguments they define to
        the context.

        Args:
            options (:class:`argparse.Namespace`): arguments that have been
                passed via the command line

        Returns:
            Dict[str, Any]: Dictionary that will be passed to Context
            initializer as kwargs.

        """
        return {"stack_names": options.stacks, "force_stacks": options.force}



"""Factories for tests."""
# pylint: disable=unused-argument
from collections import namedtuple

from mock import MagicMock

from runway.cfngin.config import Config, Stack
from runway.cfngin.context import Context

Lookup = namedtuple("Lookup", ("type", "input", "raw"))


class MockThreadingEvent(object):
    """Mock thread events."""

    def wait(self, timeout=None):
        """Mock wait method."""
        return False


class MockProviderBuilder(object):
    """Mock provider builder."""

    def __init__(self, provider, region=None):
        """Instantiate class."""
        self.provider = provider
        self.region = region

    def build(self, region=None, profile=None):
        """Mock build method."""
        return self.provider


def mock_provider(**kwargs):
    """Mock provider."""
    return MagicMock(**kwargs)


def mock_context(namespace="default", extra_config_args=None, **kwargs):
    """Mock context."""
    config_args = {"namespace": namespace}
    if extra_config_args:
        config_args.update(extra_config_args)
    config = Config(config_args)
    if kwargs.get("environment"):
        return Context(
            config=config,
            **kwargs)
    return Context(
        config=config,
        environment={},
        **kwargs)


def generate_definition(base_name, stack_id, **overrides):
    """Generate definitions."""
    definition = {
        "name": "%s.%d" % (base_name, stack_id),
        "class_path": "tests.cfngin.fixtures.mock_blueprints.%s" % (
            base_name.upper()),
        "requires": []
    }
    definition.update(overrides)
    return Stack(definition)


def mock_lookup(lookup_input, lookup_type, raw=None):
    """Mock lookup."""
    if raw is None:
        raw = "%s %s" % (lookup_type, lookup_input)
    return Lookup(type=lookup_type, input=lookup_input, raw=raw)


class SessionStub(object):
    """Stubber class for boto3 sessions made with session_cache.get_session().

    This is a helper class that should be used when trying to stub out
    get_session() calls using the boto3.stubber.

    Example Usage:

        @mock.patch('runway.cfngin.lookups.handlers.myfile.get_session',
                return_value=sessionStub(client))
        def myfile_test(self, client_stub):
            ...

    Attributes:
        client_stub (:class:`boto3.session.Session`:): boto3 session stub

    """

    def __init__(self, client_stub):
        """Instantiate class."""
        self.client_stub = client_stub

    def client(self, region):
        """Return the stubbed client object.

        Args:
            region (str): So boto3 won't complain

        Returns:
            :class:`boto3.session.Session`: The stubbed boto3 session

        """
        return self.client_stub



"""Runway context module."""
import logging
import os


LOGGER = logging.getLogger('runway')


def echo_detected_environment(env_name, env_vars):
    """Print a helper note about how the environment was determined."""
    env_override_name = 'DEPLOY_ENVIRONMENT'
    LOGGER.info("")
    if env_override_name in env_vars:
        LOGGER.info("Environment \"%s\" was determined from the %s environment variable.",
                    env_name,
                    env_override_name)
        LOGGER.info("If this is not correct, update "
                    "the value (or unset it to fall back to the name of "
                    "the current git branch or parent directory).")
    else:
        LOGGER.info("Environment \"%s\" was determined from the current "
                    "git branch or parent directory.",
                    env_name)
        LOGGER.info("If this is not the environment name, update the branch/folder name or "
                    "set an override value via the %s environment variable",
                    env_override_name)
    LOGGER.info("")


class Context(object):
    """Runway execution context."""

    env_override_name = 'DEPLOY_ENVIRONMENT'

    def __init__(self, env_name,  # pylint: disable=too-many-arguments
                 env_region, env_root, env_vars=None,
                 command=None):
        """Initialize base class."""
        self.env_name = env_name
        self.env_region = env_region
        self.env_root = env_root
        self.command = command
        self.env_vars = env_vars or os.environ.copy()
        self._env_name_from_env = bool(self.env_vars.get(self.env_override_name))

        self.echo_detected_environment()

        if not self._env_name_from_env:
            self.env_vars.update({'DEPLOY_ENVIRONMENT': self.env_name})

    def echo_detected_environment(self):
        """Print a helper note about how the environment was determined."""
        LOGGER.info("")
        if self._env_name_from_env:
            LOGGER.info("Environment \"%s\" was determined from the %s "
                        "environment variable.", self.env_name,
                        self.env_override_name)
            LOGGER.info("If this is not correct, update "
                        "the value (or unset it to fall back to the name of "
                        "the current git branch or parent directory).")
        else:
            LOGGER.info("Environment \"%s\" was determined from the current "
                        "git branch or parent directory.",
                        self.env_name)
            LOGGER.info("If this is not the environment name, update the "
                        "branch/folder name or set an override value via "
                        "the %s environment variable", self.env_override_name)
        LOGGER.info("")

    def save_existing_iam_env_vars(self):
        """Backup IAM environment variables for later restoration."""
        for i in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY',
                  'AWS_SESSION_TOKEN']:
            if i in self.env_vars:
                self.env_vars['OLD_' + i] = self.env_vars[i]

    def restore_existing_iam_env_vars(self):
        """Restore backed up IAM environment variables."""
        for i in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY',
                  'AWS_SESSION_TOKEN']:
            if 'OLD_' + i in self.env_vars:
                self.env_vars[i] = self.env_vars['OLD_' + i]
            elif i in self.env_vars:
                self.env_vars.pop(i)



"""Runway config file module."""
# pylint: disable=super-init-not-called,too-many-lines
from typing import (Any, Dict, List, Optional,  # pylint: disable=unused-import
                    Union, Iterator, TYPE_CHECKING)

# python2 supported pylint is unable to load this when in a venv
from distutils.util import strtobool  # pylint: disable=no-name-in-module,import-error
import logging
import os
import sys

from six import string_types
import yaml

from .util import MutableMap
from .variables import Variable

# python2 supported pylint sees this is cyclic even though its only for type checking
# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from .context import Context  # noqa: F401 pylint: disable=unused-import

LOGGER = logging.getLogger('runway')
NoneType = type(None)


class ConfigComponent(MutableMap):
    """Base class for Runway config components.

    Attributes:
        SUPPORTS_VARIABLES: A list of directives that support the use of
            variable.
        PRE_PROCESS_VARIABLES: A list of directives that support the use of
            variables and needs to be resolved before the component is
            processed.

    """

    SUPPORTS_VARIABLES = ['env_vars', 'environments', 'parameters']  # type: List[str]
    PRE_PROCESS_VARIABLES = []  # type: List[str]

    @property
    def data(self):
        # type: () -> Dict[str, Any]
        """Sanitized output of __dict__ with properties added.

        Removes anything that starts with ``_``.

        """
        data = {}

        for key, val in self.__dict__.items():
            if not key.startswith('_'):
                data[key] = val

        for attr in self.SUPPORTS_VARIABLES:
            data[attr] = getattr(self, attr, None)

        return data

    @property
    def env_vars(self):
        # type: () -> Any
        """Access the value of an attribute that supports variables."""
        value = self._env_vars.value  # pylint: disable=no-member
        if isinstance(value, dict):
            return value
        raise ValueError('{}.env_vars is of type {}; expected type '
                         'of dict'.format(self.name, type(value)))

    @property
    def environments(self):
        # type: () -> Dict[Any, Any]
        """Access the value of an attribute that supports variables."""
        value = self._environments.value  # pylint: disable=no-member
        if isinstance(value, dict):
            return value
        raise ValueError('{}.environments is of type {}; expected type '
                         'of dict'.format(self.name, type(value)))

    @property
    def parameters(self):
        # type: () -> Dict[str, Any]
        """Access the value of an attribute that supports variables."""
        value = self._parameters.value  # pylint: disable=no-member
        if isinstance(value, dict):
            return value
        raise ValueError('{}.parameters is of type {}; expected type '
                         'of dict'.format(self.name, type(value)))

    def get(self, key, default=None):
        # type: (str, Any) -> Any
        """Implement evaluation of get."""
        return getattr(self, key, getattr(self, key.replace('-', '_'), default))

    def resolve(self, context, variables=None, pre_process=False):
        # type: ('Context', Optional[VariablesDefinition], bool) -> None
        """Resolve attributes that support variables.

        Args:
            context: The current context object.
            variables: Object containing variables passed to Runway.
            pre_process: Only resolve the variables that are required before
                the component is processed. If this is ``False``, all variables
                will be resolved. This is useful to prevent errors when
                variables cannot be resolved because the values are not
                populated until processing has begun.

        """
        for attr in (self.PRE_PROCESS_VARIABLES if pre_process
                     else self.SUPPORTS_VARIABLES):
            LOGGER.debug('Resolving %s.%s', self.name, attr)
            getattr(self, '_' + attr).resolve(context, variables=variables)

    def __getitem__(self, key):
        # type: (str) -> Any
        """Implement evaluation of self[key]."""
        result = getattr(self, key, getattr(self, key.replace('-', '_')))

        if isinstance(result, Variable):
            return result.value
        return result

    def __len__(self):
        # type: () -> int
        """Implement the built-in function len()."""
        return len(self.data)

    def __iter__(self):
        # type: () -> Iterator[Any]
        """Return iterator object that can iterate over all attributes."""
        return iter(self.data)


class ModuleDefinition(ConfigComponent):  # pylint: disable=too-many-instance-attributes
    """A module defines the directory to be processed and applicable options.

    It can consist of `CloudFormation`_ (using `CFNgin`_),
    `Terraform`_, `Serverless Framework`_, `AWS CDK`_, `Kubernetes`_, or
    a :ref:`Static Site<mod-staticsite>`. It is recommended to place the
    appropriate extension on each directory for identification (but it
    is not required). See :ref:`Repo Structure<repo-structure>` for
    examples of a module directory structure.

    +------------------+-----------------------------------------------+
    | Suffix/Extension | IaC Tool/Framework                            |
    +==================+===============================================+
    | ``.cdk``         | `AWS CDK`_                                    |
    +------------------+-----------------------------------------------+
    | ``.cfn``         | `CloudFormation`_                             |
    +------------------+-----------------------------------------------+
    | ``.sls``         | `Serverless Framework`_                       |
    +------------------+-----------------------------------------------+
    | ``.tf``          | `Terraform`_                                  |
    +------------------+-----------------------------------------------+
    | ``.k8s``         | `Kubernetes`_                                 |
    +------------------+-----------------------------------------------+
    | ``.web``         | :ref:`Static Site<mod-staticsite>`            |
    +------------------+-----------------------------------------------+

    A module is only deployed if there is a corresponding environment file
    present or parameters are provided. This can take the form of either a file
    in the module folder or the ``parameters`` option being defined. The naming
    format varies per-module type.
    See :ref:`Module Configurations<module-configurations>` for acceptable
    environment file name formats.

    Modules can be defined as a string or a mapping. The minimum
    requirement for a module is a string that is equal to the name of
    the module directory. Providing a string is the same as providing a
    value for ``path`` in a mapping definition.

    Example:
      .. code-block:: yaml

        deployments:
          - modules:
              - my-module.cfn  # this
              - path: my-module.cfn  # is the same as this

    Using a map to define a module provides the ability to specify
    per-module ``options``, parameters, environment variables,tags,
    and even a custom class for processing the module. The options that
    can be used with each module vary. For detailed information about
    module-specific options, see
    :ref:`Module Configurations<module-configurations>`.

    Example:
      .. code-block:: yaml

        deployments:
          - modules:
              - name: my-module
                path: my-module.tf
                environments:
                  prod: 111111111111/us-east-1
                  dev:
                    - 222222222222/us-east-1
                    - 333333333333/us-east-1
                  lab: true
                parameters:
                  image_id: ${var image_id.${env DEPLOY_ENVIRONMENT}}
                tags:
                  - app:example
                  - my-tag
                options:
                  terraform_backend_config:
                    region: us-east-1
                  terraform_backend_cfn_outputs:
                    bucket: StackName::OutputName
                    dynamodb_table: StackName::OutputName

    One special map keyword, ``parallel``, indicates a list of child
    modules that will be executed in parallel (simultaneously) if the
    ``CI`` :ref:`environment variable is set<non-interactive-mode>`.

    Example:
      In this example, ``backend.tf`` will be deployed followed by the services
      that will be utilizing it. The services will be deployed in parallel.
      After the services have completed, ``frontend.tf`` will be deployed.

      .. code-block:: yaml

        deployments:
          - modules:
            - backend.tf
            - parallel:
              - servicea.cfn  # any normal module option can be used here
              - path: serviceb.cfn
              - path: servicec.cfn
                parameters:
                  count: ${var count.${env DEPLOY_ENVIRONMENT}}
            - frontend.tf

    """

    SUPPORTS_VARIABLES = ['class_path', 'env_vars', 'environments',
                          'options', 'parameters', 'path']

    def __init__(self,  # pylint: disable=too-many-arguments
                 name,  # type: str
                 path,  # type: str
                 class_path=None,  # type: Optional[str]
                 type_str=None,  # type: Optional[str]
                 environments=None,  # type: Optional[Dict[str, Dict[str, Any]]]
                 parameters=None,  # type: Optional[Dict[str, Any]]
                 env_vars=None,  # type: Optional[Dict[str, Dict[str, Any]]]
                 options=None,  # type: Optional[Dict[str, Any]]
                 tags=None,  # type: Optional[Dict[str, str]]
                 child_modules=None  # type: Optional[List[Union[str, Dict[str, Any]]]]
                 # pylint only complains for python2
                 ):  # pylint: disable=bad-continuation
        # type: (...) -> None
        """.. Runway module definition.

        Keyword Args:
            name (str): Name of the module. Used to more easily identify
                where different modules begin/end in the logs.
            path (str): Path to the module relative to the Runway config
                file. This cannot be higher than the Runway config file. See
                `Path`_ for detailed usage.
            class_path (Optional[str]): Path to custom Runway module class.
                Also used for static site deployments. See
                :ref:`Module Configurations<module-configurations>` for
                detailed usage.
            type_str (Optional[str]): Alias for type of module to use
                :ref:`Module Configurations<module-configurations>` for
                detailed usage.
            environments (Optional[Dict[str, Dict[str, Any]]]): Optional
                mapping of environment names to a booleon value used to
                explicitly deploy or not deploy in an environment. This
                can be used when an environment specific variables file
                and parameters are not needed to force a module to deploy
                anyway or, explicitly skip a module even if a file or
                parameters are found. The mapping can also have a string
                (or list of strings) value of $ACCOUNT_ID/$REGION to lock
                an environment to specific regions in a specific accounts.
                If it matches, it will act as an explicit deploy.
            env_vars (Optional[Dict[str, Dict[str, Any]]]): A mapping of
                OS environment variable overrides to apply when processing
                modules in the deployment. Can be defined per environment
                or for all environments by omiting the environment name.
                Takes precedence over values set at the deployment-level.
            options (Optional[Dict[str, Any]]): Module-specific options.
                See :ref:`Module Configurations<module-configurations>`
                for detailed usage. Takes precedence over values set at the
                deployment-level.
            parameters (Optional(Dict[str, Any])): Module level parameters that
                are akin to a `CloudFormation`_ parameter in functionality.
                These can be used to pass variable values to your modules in
                place of a ``.env``/``.tfenv``/environment config file.
                Through the use of `Lookups`_, the value can differ per
                deploy environment, region, etc.
            tags (Optional[Dict[str, str]]): Module tags used to select
                which modules to process using CLI arguments.
                (``--tag <tag>...``)
            child_modules (Optional[List[Union[str, Dict[str, Any]]]]):
                Child modules that can be executed in parallel

        .. rubric:: Lookup Resolution

        +---------------------+-----------------------------------------------+
        | Keyword / Directive | Support                                       |
        +=====================+===============================================+
        |  ``name``           | None                                          |
        +---------------------+-----------------------------------------------+
        |  ``path``           | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``class_path``     | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``environments``   | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``env_vars``       | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``options``        | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``parameters``     | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``tags``           | None                                          |
        +---------------------+-----------------------------------------------+

        References:
            - `AWS CDK`_
            - `CloudFormation`_
            - `Serverless Framework`_
            - `CFNgin`_
            - `Troposphere`_
            - `Terraform`_
            - `Kubernetes`_
            - :ref:`Static Site<mod-staticsite>`
            - :ref:`Module Configurations<module-configurations>` -
              detailed module ``options``
            - :ref:`Repo Structure<repo-structure>` - examples of
              directory structure
            - :ref:`command-deploy`
            - :ref:`command-destroy`
            - :ref:`command-plan`

        """
        self.name = name
        self._path = Variable(name + '.path', path, 'runway')
        self._class_path = Variable(name + '.class_path', class_path, 'runway')
        self.type = type_str
        self._environments = Variable(name + '.environments',
                                      environments or {}, 'runway')
        self._parameters = Variable(name + '.parameters', parameters or {},
                                    'runway')
        self._env_vars = Variable(name + '.env_vars', env_vars or {}, 'runway')
        self._options = Variable(name + '.options', options or {}, 'runway')
        self.tags = tags or {}
        self.child_modules = child_modules or []

    @property
    def class_path(self):
        # type: () -> Optional[str]
        """Access the value of an attribute that supports variables."""
        value = self._class_path.value
        if not value:
            return None
        if isinstance(value, str):
            return value
        raise ValueError('{}.class_path = {} is of type {}; expected type '
                         'of str'.format(self.name, value, type(value)))

    @property
    def options(self):
        # type: () -> Dict[Any, Any]
        """Access the value of an attribute that supports variables."""
        value = self._options.value
        if isinstance(value, dict):
            return value
        raise ValueError('{}.options is of type {}; expected type '
                         'of dict'.format(self.name, type(value)))

    @property
    def path(self):
        # type: () -> str
        """Access the value of an attribute that supports variables."""
        value = self._path.value  # pylint: disable=no-member
        if isinstance(value, str):
            return value
        raise ValueError('{}.path is of type {}; expected type '
                         'of str'.format(self.name, type(value)))

    @classmethod
    def from_list(cls, modules):
        # type: (List[Union[Dict[str, Any], str]]) -> List[ModuleDefinition]
        """Instantiate ModuleDefinition from a list."""
        results = []

        for mod in modules:
            if isinstance(mod, str):
                results.append(cls(name=mod, path=mod))
                continue
            if mod.get('parallel'):
                name = 'parallel_parent'
                child_modules = ModuleDefinition.from_list(mod.pop('parallel'))
                path = '[' + ', '.join([x.path for x in child_modules]) + ']'
                if mod:
                    LOGGER.warning(
                        'Invalid keys found in parallel module config have been ignored: %s',
                        ', '.join(mod.keys())
                    )
            else:
                name = mod.pop('name', mod['path'])
                child_modules = None
                path = mod.pop('path')
            results.append(cls(name,
                               path,
                               class_path=mod.pop('class_path', None),
                               type_str=mod.pop('type', None),
                               environments=mod.pop('environments', {}),
                               env_vars=mod.pop('env_vars', {}),
                               options=mod.pop('options', {}),
                               parameters=mod.pop('parameters', {}),
                               tags=mod.pop('tags', {}),
                               child_modules=child_modules))
            if mod:
                LOGGER.warning(
                    'Invalid keys found in module %s have been ignored: %s',
                    name, ', '.join(mod.keys())
                )
        return results


class DeploymentDefinition(ConfigComponent):  # pylint: disable=too-many-instance-attributes
    """A deployment defines modules and options that affect the modules.

    Deployments are processed during a ``deploy``/``destroy``/``plan``
    action. If the processing of one deployment fails, the action will
    end.

    During a ``deploy``/``destroy`` action, the user has the option to
    select which deployment will run unless the ``CI`` environment
    variable is set, the ``--tag <tag>...`` cli option was provided, or
    only one deployment is defined.

    Example:
      .. code-block:: yaml

        deployments:
          - modules:  # minimum requirements for a deployment
              # "./" can alternatively be used for the module name to indicate
              # the current directory
              - my-module.cfn
            regions:
              - us-east-1
          - name: detailed-deployment  # optional
            modules:
              - path: my-other-modules.cfn
                type: cloudformation
            regions:
              - us-east-1
            environments:
              prod: 111111111111/us-east-1
              dev:
                - 222222222222/us-east-1
                - 333333333333/us-east-1
              lab: true
            account_id: ${var account_ids}  # optional
            assume_role: ${var assume_role}  # optional
            parameters:  # optional
                region: ${env AWS_REGION}
                image_id: ${var image_id.${env DEPLOY_ENVIRONMENT}}
            env_vars:  # optional environment variable overrides
                AWS_PROFILE: ${var aws_profile.${env DEPLOY_ENVIRONMENT}::default=default}
                APP_PATH: ${var app_path.${env DEPLOY_ENVIRONMENT}}

    """

    SUPPORTS_VARIABLES = ['account_alias', 'account_id', 'assume_role',
                          'env_vars', 'environments', 'module_options',
                          'regions', 'parallel_regions', 'parameters']
    PRE_PROCESS_VARIABLES = ['account_alias', 'account_id', 'assume_role',
                             'env_vars', 'regions']

    def __init__(self, deployment):
        # type: (Dict[str, Any]) -> None
        """.. Runway deployment definition.

        Keyword Args:
            account_alias (Optional[Dict[str, str]]): A mapping of
                ``$environment: $alias`` that, if provided, is used to
                verify the currently assumed role or credentials.
            account_id (Optional[Dict[str, Union[str, int]]]): A mapping
                of ``$environment: $id`` that, if provided, is used to
                verify the currently assumed role or credentials.
            assume_role (Optional[Dict[str, Union[str, Dict[str, str]]]]):
                A mapping of ``$environment: $role`` or
                ``$environment: {arn: $role, duration: $int}`` to assume
                a role when processing a deployment. ``arn: $role`` can
                be used to apply the same role to all environment.
                ``post_deploy_env_revert: true`` can also be provided to
                revert credentials after processing.
            environments (Optional[Dict[str, Dict[str, Any]]]): Optional
                mapping of environment names to a booleon value used to
                explicitly enable or disable in an environment. This
                can be used when an environment specific variables file
                and parameters are not needed to force a module to enable
                anyway or, explicitly skip a module even if a file or
                parameters are found. The mapping can also have a string
                (or list of strings) value of $ACCOUNT_ID/$REGION to lock
                an environment to specific regions in a specific accounts.
                If it matches, it will act as an explicit enable.
            env_vars (Optional[Dict[str, Dict[str, Any]]]): A mapping of
                OS environment variable overrides to apply when processing
                modules in the deployment. Can be defined per environment
                or for all environments by omiting the environment name.
            modules (Optional[List[Dict[str, Any]]]): A list of modules
                to be processed in the order they are defined.
            module_options (Optional[Dict[str, Any]]): Options that are
                shared among all modules in the deployment.
            name (str): Name of the deployment. Used to more easily
                identify where different deployments begin/end in the logs.
            type (str): The type of module we are deploying. By default
                Runway will first check to see if you explicitly specify
                the module type, after that it will check to see if a
                valid module extension exists on the directory, and
                finally it will attempt to autodetect the type of module.
                Valid values are: ``serverless``, ``terraform``, ``cdk``,
                ``kubernetes``, ``cloudformation``, ``static``.
            regions (List[str]): AWS region names where modules will be
                deployed/destroyed. Can optionally define as a map with
                ``parallel`` as the key and a list of regions as the value.
                See **parallel_regions** for more info.
            parallel_regions: Can be defined in place of ``regions.parallel[]``.
                This will cause all modules in the deployment to be executed
                in all provided regions in parallel (at the same time).
                Only takes effect when the ``CI`` environment variable is set,
                enabling non-interactive mode, as prompts will not be able
                to be presented. If ``CI`` is not set, the regions will be
                processed one at a time. This can be used in tandom with
                **parallel modules**. ``assume_role.post_deploy_env_revert``
                will always be ``true`` when run in parallel.
            parameters (Optional(Dict[str, Any])): Module level parameters that
                are akin to a `CloudFormation`_ parameter in functionality.
                These can be used to pass variable values to your modules in
                place of a ``.env``/``.tfenv``/environment config file.
                Through the use of `Lookups`_, the value can differ per
                deploy environment, region, etc.

        .. rubric:: Lookup Resolution

        .. important:: Due to how a deployment is processed, values are
                       resolved twice. Once before processing and once during
                       processing. Because of this, the keywords/directives
                       that are resolved before processing will not have
                       access to values set during process like ``AWS_REGION``,
                       ``AWS_DEFAULT_REGION``, and ``DEPLOY_ENVIRONMENT`` for
                       the pre-processing resolution but, if they are resolved
                       again during processing, these will be available. To
                       avoide errors during the first resolution due to the
                       value not existing, provide a default value for the
                       :ref:`Lookup <Lookups>`.

        +---------------------+-----------------------------------------------+
        | Keyword / Directive | Support                                       |
        +=====================+===============================================+
        | ``account_alias``   | `env lookup`_ (``AWS_REGION`` and             |
        |                     | ``AWS_DEFAULT_REGION`` will not have been set |
        |                     | by Runway yet), `var lookup`_                 |
        +---------------------+-----------------------------------------------+
        | ``account_id``      | `env lookup`_ (``AWS_REGION`` and             |
        |                     | ``AWS_DEFAULT_REGION`` will not have been set |
        |                     | by Runway yet), `var lookup`_                 |
        +---------------------+-----------------------------------------------+
        |  ``assume_role``    | `env lookup`_ (``AWS_REGION`` and             |
        |                     | ``AWS_DEFAULT_REGION`` will not have been set |
        |                     | by Runway yet), `var lookup`_                 |
        +---------------------+-----------------------------------------------+
        |  ``environments``   | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``env_vars``       | `env lookup`_ (``AWS_REGION``,                |
        |                     | ``DEPLOY_ENVIRONMENT``, and                   |
        |                     | ``AWS_DEFAULT_REGION`` will not have been set |
        |                     | by Runway during pre-process resolution.      |
        |                     | provide a default value to avoide errors.),   |
        |                     | `var lookup`_                                 |
        +---------------------+-----------------------------------------------+
        |  ``modules``        | No direct support. See `module`_ for details  |
        |                     | on support within a module definition.        |
        +---------------------+-----------------------------------------------+
        |  ``module_options`` | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        |  ``name``           | None                                          |
        +---------------------+-----------------------------------------------+
        |  ``regions``        | `env lookup`_ (``AWS_REGION`` and             |
        |                     | ``AWS_DEFAULT_REGION`` will not have been set |
        |                     | by Runway yet), `var lookup`_                 |
        +---------------------+-----------------------------------------------+
        | ``parallel_regions``| `env lookup`_ (``AWS_REGION`` and             |
        |                     | ``AWS_DEFAULT_REGION`` will not have been set |
        |                     | by Runway yet), `var lookup`_                 |
        +---------------------+-----------------------------------------------+
        |  ``parameters``     | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+

        References:
            - :class:`module<runway.config.ModuleDefinition>`
            - :ref:`command-deploy`
            - :ref:`command-destroy`
            - :ref:`command-plan`

        """
        self._reverse = False
        self.name = deployment.pop('name')  # type: str
        self._account_alias = Variable(
            self.name + '.account_alias', deployment.pop(
                'account_alias', deployment.pop('account-alias', {})
            ), 'runway'
        )  # type: Variable
        self._account_id = Variable(self.name + '.account_id', deployment.pop(
            'account_id', deployment.pop('account-id', {})
        ), 'runway')  # type: Variable
        self._assume_role = Variable(
            self.name + '.assume_role', deployment.pop(
                'assume_role', deployment.pop('assume-role', {})
            ), 'runway'
        )  # type: Variable
        self._environments = Variable(
            self.name + '.environments', deployment.pop('environments', {}),
            'runway'
        )  # type: Variable
        self._parameters = Variable(
            self.name + '.parameters', deployment.pop('parameters', {}),
            'runway'
        )  # type: Variable
        self._env_vars = Variable(self.name + '.env_vars', deployment.pop(
            'env_vars', deployment.pop('env-vars', {})
        ), 'runway')  # type: Variable
        if deployment.pop('current_dir', False):
            # Deprecated in 1.0 (late 2019). Retain for at least a major version.
            LOGGER.warning('DEPRECATION WARNING: The "current_dir" option has '
                           'been deprecated in favor of a "./" module '
                           'definition. Please update your config.')
            modules = ['.' + os.sep]
        else:
            if not deployment.get('modules'):
                LOGGER.error('No modules have been defined in your Runway '
                             'deployment.')
                sys.exit(1)
            modules = deployment.pop('modules')
        self.modules = ModuleDefinition.from_list(
            modules
        )  # type: List[ModuleDefinition]
        self._module_options = Variable(
            self.name + '.module_options', deployment.pop(
                'module_options', deployment.pop('module-options', {})
            ), 'runway'
        )  # type: Variable

        # should add variable resolve here to support parallel region
        # dict in the resolved variable
        regions = deployment.pop(
            'regions', []
        )

        if regions and deployment.get('parallel_regions'):
            LOGGER.error('Found "regions" and "parallel_regions" in '
                         'deployment "%s"; only one can be defined',
                         self.name)
            sys.exit(1)
        if isinstance(regions, dict) and regions.get('parallel'):
            self._parallel_regions = Variable(
                self.name + '.parallel_regions', regions.pop('parallel'),
                'runway'
            )  # type: Variable
            self._regions = Variable(self.name + '.regions',
                                     [], 'runway')  # type: Variable
        else:
            self._regions = Variable(self.name + '.regions',
                                     regions, 'runway')
            self._parallel_regions = Variable(
                self.name + '.parallel_regions',
                deployment.pop('parallel_regions', []),
                'runway'
            )

        if deployment:
            LOGGER.warning(
                'Invalid keys found in deployment %s have been ignored: %s',
                self.name, ', '.join(deployment.keys())
            )

    @property
    def account_alias(self):
        # type: () -> Union[Dict[Any, Any], str]
        """Access the value of an attribute that supports variables."""
        value = self._account_alias.value
        if isinstance(value, (dict, string_types)):
            return value
        raise ValueError('{}.account_alias is of type {}; expected type '
                         'of dict or str'.format(self.name, type(value)))

    @property
    def account_id(self):
        # type: () -> Union[Dict[Any, Any], str]
        """Access the value of an attribute that supports variables."""
        value = self._account_id.value
        if isinstance(value, (dict, string_types)):
            return value
        if isinstance(value, int):
            return str(value)
        raise ValueError('{}.account_id is of type {}; expected type '
                         'of dict, int, or str'.format(self.name, type(value)))

    @property
    def assume_role(self):
        # type: () -> Union[Dict[Any, Any], str]
        """Access the value of an attribute that supports variables."""
        value = self._assume_role.value
        if isinstance(value, (dict, string_types)):
            return value
        raise ValueError('{}.assume_role is of type {}; expected type '
                         'of dict or str'.format(self.name, type(value)))

    @property
    def module_options(self):
        # type: () -> Dict[Any, Any]
        """Access the value of an attribute that supports variables."""
        value = self._module_options.value
        if isinstance(value, dict):
            return value
        raise ValueError('{}.module_options is of type {}; expected type '
                         'of dict'.format(self.name, type(value)))

    @property
    def regions(self):
        # type: () -> List[str]
        """Access the value of an attribute that supports variables."""
        value = self._regions.value
        if isinstance(value, list):
            if self._reverse:
                return value[::-1]
            return value
        raise ValueError('{}.regions is of type {}; expected type '
                         'of list'.format(self.name, type(value)))

    @property
    def parallel_regions(self):
        # type: () -> List[str]
        """Access the value of an attribute that supports variables."""
        value = self._parallel_regions.value
        if isinstance(value, list):
            return value
        raise ValueError('{}.parallel_regions is of type {}; expected type '
                         'of list'.format(self.name, type(value)))

    def reverse(self):
        """Reverse the order of modules and regions."""
        if self._reverse:
            self._reverse = False
        else:
            self._reverse = True
        self.modules.reverse()

    @classmethod
    def from_list(cls, deployments):
        # type: (Optional[List[Dict[str, Any]]]) -> List[DeploymentDefinition]
        """Instantiate DeploymentDefinitions from a list."""
        results = []

        if not deployments:
            return []

        for i, deployment in enumerate(deployments):
            if not deployment.get('name'):
                deployment['name'] = 'deployment_{}'.format(str(i + 1))
            results.append(cls(deployment))
        return results


class TestDefinition(ConfigComponent):
    """Tests can be defined as part of the Runway config file.

    This is to remove the need for complex Makefiles or scripts to initiate
    test runners. Simply define all tests for a project in the Runway
    config file and use the ``runway test`` :ref:`command<command-test>`
    to execute them.

    Example:
      .. code-block:: yaml

        tests:
          - name: my-test
            type: script
            required: false
            args:
              commands:
                - echo "Hello World!"

    """

    SUPPORTS_VARIABLES = ['args', 'required']

    def __init__(self,
                 name,  # type: str
                 test_type,  # type: str
                 args=None,  # type: Optional[Dict[str, Any]]
                 required=True  # type: bool
                 # pylint only complains for python2
                 ):  # pylint: disable=bad-continuation
        # type: (...) -> None
        """.. Runway test definitions.

        Keyword Args:
            name (str): Name of the test. Used to more easily identify
                where different tests begin/end in the logs.
            type (str): The type of test to run. See
                :ref:`Build-in Test Types<built-in-test-types>`
                for supported test types.
            args (Optional[Dict[str, Any]]): Arguments to be passed to
                the test. Supported arguments vary by test type. See
                :ref:`Build-in Test Types<built-in-test-types>` for the
                list of arguments supported by each test type.
            required (bool):  If false, testing will continue if the test
                fails. *(default: true)*

        .. rubric:: Lookup Resolution

        .. note:: Runway does not set ``AWS_REGION`` or ``AWS_DEFAULT_REGION``
                  environment variables. If the ``DEPLOY_ENVIRONMENT``
                  environment variable is not manually set, it will always
                  be ``test`` and is not determined from the branch or
                  directory.

        +---------------------+-----------------------------------------------+
        | Keyword / Directive | Support                                       |
        +=====================+===============================================+
        | ``args``            | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+
        | ``required``        | `env lookup`_, `var lookup`_                  |
        +---------------------+-----------------------------------------------+

        References:
            - :ref:`Build-in Test Types<built-in-test-types>` - Supported
              test types and their
              arguments
            - :ref:`test command<command-test>`

        """
        self.name = name
        self.type = test_type
        self._args = Variable(self.name + '.args', args or {}, 'runway')
        self._required = Variable(self.name + '.required', required, 'runway')

    @property
    def args(self):
        # type: () -> Dict[str, Any]
        """Access the value of an attribute that supports variables."""
        value = self._args.value
        if isinstance(value, dict):
            return value
        raise ValueError('{}.args is of type {}; expected type '
                         'of dict'.format(self.name, type(value)))

    @property
    def required(self):
        # type: () -> bool
        """Access the value of an attribute that supports variables."""
        value = self._required.value
        if isinstance(value, bool):
            return value
        try:
            value = strtobool(value)
            return value
        except ValueError:
            pass
        raise ValueError('{}.required is of type {}; expected type '
                         'of bool'.format(self.name, type(value)))

    @classmethod
    def from_list(cls, tests):
        # type: (Optional[List[Dict[str, Any]]]) -> List[TestDefinition]
        """Instantiate TestDefinitions from a list."""
        results = []

        if not tests:
            return []

        for index, test in enumerate(tests):
            name = test.pop('name', 'test_{}'.format(index + 1))
            results.append(cls(name, test.pop('type'),
                               test.pop('args', {}),
                               test.pop('required', False)))

            if test:
                LOGGER.warning(
                    'Invalid keys found in test %s have been ignored: %s',
                    name, ', '.join(test.keys())
                )
        return results


class VariablesDefinition(MutableMap):
    """A variable definitions for the Runway config file.

    Runway variables are used to fill values that could change based on any
    number of circumstances. They can also be used to simplify the Runway config
    file by pulling lengthy definitions into another file. Variables can be used
    in the config file by providing the `var lookup`_ to any keyword/directive
    that supports :ref:`Lookups <Lookups>`.

    By default, Runway will look for and load a ``runway.variables.yml`` or
    ``runway.variables.yaml`` file that is in the same directory as the
    Runway config file. The file path and name of the file can optionally be
    defined in the config file. If the file path is explicitly provided and
    the file can't be found, an error will be raised.

    Variables can also be defined in the Runway config file directly. This can
    either be in place of a dedicated variables file, extend an existing file,
    or override values from the file.

    .. rubric:: Lookup Resolution

    Runway lookup resolution is not supported within the variables definition
    block or variables file. Attempts to use Runway :ref:`Lookups <Lookups>`
    within the variables definition block or variables file will result in
    the literal value being processed.

    Example:
      .. code-block:: yaml

        variables:
          sys_path: ./  # defaults to the current directory
          file_path: secrets.yaml
          # define additional variables or override those in the variables file
          another_var: some_value
        deployments:
          - modules:
              - ${var sampleapp.definition}
            regions: ${var sampleapp.regions}

    """

    default_names = ['runway.variables.yml', 'runway.variables.yaml']

    def __init__(self, file_path=None, sys_path=None, **kwargs):
        """.. Not really needed but cleans up the docs.

        Keyword Args:
            file_path: Explicit path to a variables file. If it cannot be found
                Runway will exit.
            sys_path: Directory to base relative paths off of.

        """
        self._file_path = file_path
        self._sys_path = sys_path
        super(VariablesDefinition, self).__init__(**kwargs)

    @classmethod
    def find_file(cls, file_path=None, sys_path=None):
        # type: (Any, Any) -> Optional[str]
        """Find a Runway variables file.

        Args:
            file_path: Explicit path to a variables file. If it cannot be found
                Runway will exit.
            sys_path: Directory to base relative paths off of.

        Returns:
            Verified path to a file.

        Raises:
            TypeError: file_path or sys_path is not a string.

        """
        if not (isinstance(file_path, (str, NoneType)) and
                isinstance(file_path, (str, NoneType))):
            raise TypeError('file_path and sys_path of VariablesDefinition '
                            'must of be of type str but got types {} and {}'.format(
                                type(file_path), type(sys_path)))
        if not sys_path:
            sys_path = os.getcwd()

        if file_path:
            result = os.path.join(sys_path, file_path)
            if os.path.isfile(result):
                return result
            LOGGER.error('The provided variables "%s" file could not '
                         'be found.', result)
            sys.exit(1)

        for name in cls.default_names:
            result = os.path.join(sys_path, name)
            if os.path.isfile(result):
                return result

        LOGGER.info('Could not find %s in the current directory. '
                    'Continuing without a variables file.',
                    ' or '.join(cls.default_names))
        return None

    @classmethod
    def load(cls, **kwargs):
        # type: (Dict[str, Any]) -> VariablesDefinition
        """Load variables."""
        file_path = cls.find_file(file_path=kwargs.pop('file_path', None),
                                  sys_path=kwargs.pop('sys_path', None))

        if file_path:
            variables = cls._load_from_file(file_path)

            for key, val in kwargs.items():
                variables[key] = val

            return variables

        return cls(**kwargs)

    @classmethod
    def _load_from_file(cls, file_path):
        # type: (str) -> VariablesDefinition
        """Load the variables file into an object."""
        if not os.path.isfile(file_path):
            LOGGER.error('The provided variables "%s" file could not '
                         'be found.', file_path)
            sys.exit(1)

        with open(file_path) as data_file:
            return cls(**yaml.safe_load(data_file))


class Config(ConfigComponent):
    """The Runway config file is where all options are defined.

    It contains definitions for deployments, tests, and some global
    options that impact core functionality.

    The Runway config file can have two possible names, ``runway.yml``
    or ``runway.yaml``. It must be stored at the root of the directory
    containing all modules to be deployed.

    Example:
        .. code-block:: yaml

            ---
            # See full syntax at https://github.com/onicagroup/runway
            ignore_git_branch: true
            tests:
              - name: example
                type: script
                args:
                  commands:
                    - echo "Hello world"
            deployments:
              - modules:
                  - path: my-modules.cfn
                regions:
                  - us-east-1

    """

    accepted_names = ['runway.yml', 'runway.yaml']

    def __init__(self,
                 deployments,  # type: List[Dict[str, Any]]
                 tests=None,  # type: List[Dict[str, Any]]
                 ignore_git_branch=False,  # type: bool
                 variables=None  # type: Optional[Dict[str, Any]]
                 # pylint only complains for python2
                 ):  # pylint: disable=bad-continuation
        # type: (...) -> None
        """.. Top-level Runway config file.

        Keyword Args:
            deployments (List[Dict[str, Any]]): A list of
                :class:`deployments<runway.config.DeploymentDefinition>`
                that are processed in the order they are defined.
            tests (Optional[List[Dict[str, Any]]]): A list of
                :class:`tests<runway.config.TestDefinition>` that are
                processed in the order they are defined.
            ignore_git_branch (bool): Disable git branch lookup when
                using environment folders, non-git VCS, or defining the
                ``DEPLOY_ENVIRONMENT`` environment variable before
                execution. Note that defining ``DEPLOY_ENVIRONMENT``
                will automatically ignore the git branch.
            variables (Optional[Dict[str, Any]]): A map that defines the
                location of a variables file and/or the variables
                themselves.

        .. rubric:: Lookup Resolution

        +---------------------+-----------------------------------------------+
        | Keyword / Directive | Support                                       |
        +=====================+===============================================+
        | ``deployments``     | No direct support. See `Deployment`_ for      |
        |                     | details on support within a deploymet         |
        |                     | definition.                                   |
        +---------------------+-----------------------------------------------+
        | ``tests``           | No direct support. See `Test`_ for details on |
        |                     | support within a test definition.             |
        +---------------------+-----------------------------------------------+
        |``ignore_git_branch``| None                                          |
        +---------------------+-----------------------------------------------+
        | ``variables``       | None                                          |
        +---------------------+-----------------------------------------------+

        References:
            - :class:`deployment<runway.config.DeploymentDefinition>`
            - :class:`test<runway.config.TestDefinition>`

        """
        self.deployments = DeploymentDefinition.from_list(deployments)
        self.tests = TestDefinition.from_list(tests)
        self.ignore_git_branch = ignore_git_branch

        variables = variables or {}
        self.variables = VariablesDefinition.load(**variables)

    @classmethod
    def load_from_file(cls, config_path):
        # type: (str) -> Config
        """Load config file into a Config object."""
        if not os.path.isfile(config_path):
            LOGGER.error("Runway config file was not found (looking for "
                         "%s)",
                         config_path)
            sys.exit(1)
        with open(config_path) as data_file:
            config_file = yaml.safe_load(data_file)
            result = Config(config_file.pop('deployments'),
                            config_file.pop('tests', []),
                            config_file.pop('ignore_git_branch',
                                            config_file.pop(
                                                'ignore-git-branch',
                                                False)))

            if config_file:
                LOGGER.warning(
                    'Invalid keys found in runway file have been ignored: %s',
                    ', '.join(config_file.keys())
                )
            return result

    @classmethod
    def find_config_file(cls, config_dir=None):
        # type: (Optional[str]) -> str
        """Find the Runway config file."""
        if not config_dir:
            config_dir = os.getcwd()

        for name in cls.accepted_names:
            conf_path = os.path.join(config_dir, name)
            if os.path.isfile(conf_path):
                return conf_path

        LOGGER.error('Runway config file was not found. Looking for one '
                     'of %s in %s', str(cls.accepted_names), config_dir)
        sys.exit(1)



"""CFNgin plan, plan componenets, and functions for interacting with a plan."""
import json
import logging
import os
import threading
import time
import uuid

from .dag import DAG, DAGValidationError, walk
from .exceptions import (CancelExecution, GraphError, PersistentGraphLocked,
                         PlanFailed)
from .status import (COMPLETE, FAILED, PENDING, SKIPPED, SUBMITTED,
                     FailedStatus, SkippedStatus)
from .ui import ui
from .util import merge_map, stack_template_key_name

LOGGER = logging.getLogger(__name__)

COLOR_CODES = {
    SUBMITTED.code: 33,  # yellow
    COMPLETE.code: 32,   # green
    FAILED.code: 31,     # red
}


def json_serial(obj):
    """Serialize json.

    Args:
        obj (Any): A python object.

    Example:
        json.dumps(data, default=json_serial)

    """
    if isinstance(obj, set):
        return list(obj)
    raise TypeError


def log_step(step):
    """Construct a log message for a set and log it to the UI.

    Args:
        step (:class:`Step`): The step to be logged.

    """
    msg = "%s: %s" % (step, step.status.name)
    if step.status.reason:
        msg += " (%s)" % (step.status.reason)
    color_code = COLOR_CODES.get(step.status.code, 37)
    ui.info(msg, extra={"color": color_code})


def merge_graphs(graph1, graph2):
    """Combine two Graphs into one, retaining steps.

    Args:
        graph1 (:class:`Graph`): Graph that ``graph2`` will
            be merged into.
        graph2 (:class:`Graph`): Graph that will be merged
            into ``graph1``.

    Returns:
        :class:`Graph`: A combined graph.

    """
    merged_graph_dict = merge_map(graph1.to_dict().copy(),
                                  graph2.to_dict())
    steps = [graph1.steps.get(name, graph2.steps.get(name))
             for name in merged_graph_dict.keys()]
    return Graph.from_steps(steps)


class Step(object):
    """State machine for executing generic actions related to stacks.

    Attributes:
        fn (Optional[Callable]): Function to run to execute the step.
            This function will be ran multiple times until the step is "done".
        last_updated (float): Time when the step was last updated.
        stack (:class:`runway.cfngin.stack.Stack`): the stack associated with
            this step
        status (:class:`runway.cfngin.status.Status`): The status of step.
        watch_func (Optional[Callable]): Function that will be called to
            "tail" the step action.

    """

    def __init__(self, stack, fn=None, watch_func=None):
        """Instantiate class.

        Args:
            stack (:class:`runway.cfngin.stack.Stack`): The stack associated
                with this step
            fn (Optional[Callable]): Function to run to execute the step.
                This function will be ran multiple times until the step is
                "done".
            watch_func (Optional[Callable]): Function that will be called to
                "tail" the step action.

        """
        self.stack = stack
        self.status = PENDING
        self.last_updated = time.time()
        self.fn = fn
        self.watch_func = watch_func

    def run(self):
        """Run this step until it has completed or been skipped.

        Returns:
            bool

        """
        stop_watcher = threading.Event()
        watcher = None
        if self.watch_func:
            watcher = threading.Thread(
                target=self.watch_func,
                args=(self.stack, stop_watcher)
            )
            watcher.start()

        try:
            while not self.done:
                self._run_once()
        finally:
            if watcher:
                stop_watcher.set()
                watcher.join()
        return self.ok

    def _run_once(self):
        """Run a step exactly once.

        Returns:
            str

        """
        try:
            status = self.fn(self.stack, status=self.status)
        except CancelExecution:
            status = SkippedStatus('canceled execution')
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.exception(err)
            status = FailedStatus(reason=str(err))
        self.set_status(status)
        return status

    @property
    def name(self):
        """Name of the step.

        This is equal to the name of the stack it operates on.

        Returns:
            str

        """
        return self.stack.name

    @property
    def requires(self):
        """Return a list of step names this step depends on.

        Returns:
            List[str]

        """
        return self.stack.requires

    @property
    def required_by(self):
        """Return a list of step names that depend on this step.

        Returns:
            List[str]

        """
        return self.stack.required_by

    @property
    def completed(self):
        """Return True if the step is in a COMPLETE state.

        Returns:
            bool

        """
        return self.status == COMPLETE

    @property
    def skipped(self):
        """Return True if the step is in a SKIPPED state.

        Returns:
            bool

        """
        return self.status == SKIPPED

    @property
    def failed(self):
        """Return True if the step is in a FAILED state.

        Returns:
            bool

        """
        return self.status == FAILED

    @property
    def done(self):
        """Return True if the step is finished.

        To be ``True``, status must be either COMPLETE, SKIPPED or FAILED)

        Returns:
            bool

        """
        return self.completed or self.skipped or self.failed

    @property
    def ok(self):
        """Return True if the step is finished (either COMPLETE or SKIPPED).

        Returns:
            bool

        """
        return self.completed or self.skipped

    @property
    def submitted(self):
        """Return True if the step is SUBMITTED, COMPLETE, or SKIPPED.

        Returns:
            bool

        """
        return self.status >= SUBMITTED

    def set_status(self, status):
        """Set the current step's status.

        Args:
            status (:class:`runway.cfngin.status.Status`): The status to set the
                step to.

        """
        if status is not self.status:
            LOGGER.debug("Setting %s state to %s.", self.stack.name,
                         status.name)
            self.status = status
            self.last_updated = time.time()
            if self.stack.logging:
                log_step(self)

    def complete(self):
        """Shortcut for ``set_status(COMPLETE)``."""
        self.set_status(COMPLETE)

    def skip(self):
        """Shortcut for ``set_status(SKIPPED)``."""
        self.set_status(SKIPPED)

    def submit(self):
        """Shortcut for ``set_status(SUBMITTED)``."""
        self.set_status(SUBMITTED)

    @classmethod
    def from_stack_name(cls, stack_name, context, requires=None, fn=None,
                        watch_func=None):
        """Create a step using only a stack name.

        Args:
            stack_name (str): Name of a CloudFormation stack.
            context (:class:`runway.cfngin.context.Context`): Context object.
                Required to initialize a "fake" :class:`runway.cfngin.stack.Stack`.
            requires (List[str]): Stacks that this stack depends on.
            fn (Callable): The function to run to execute the step.
                This function will be ran multiple times until the step
                is "done".
            watch_func (Callable): an optional function that will be
                called to "tail" the step action.

        Returns:
            :class:`Step`

        """
        from runway.cfngin.config import Stack as StackConfig
        from runway.cfngin.stack import Stack

        stack_def = StackConfig({'name': stack_name,
                                 'requires': requires or []})
        stack = Stack(stack_def, context)
        return cls(stack, fn=fn, watch_func=watch_func)

    @classmethod
    def from_persistent_graph(cls, graph_dict, context, fn=None,
                              watch_func=None):
        """Create a steps for a persistent graph dict.

        Args:
            graph_dict (Dict[str, List[str]]): A graph dict.
            context (:class:`runway.cfngin.context.Context`): Context object.
                Required to initialize a "fake" :class:`runway.cfngin.stack.Stack`.
            requires (List[str]): Stacks that this stack depends on.
            fn (Callable): The function to run to execute the step.
                This function will be ran multiple times until the step
                is "done".
            watch_func (Callable): an optional function that will be
                called to "tail" the step action.

        Returns:
            List[:class:`Step`]

        """
        steps = []

        for name, requires in graph_dict.items():
            steps.append(cls.from_stack_name(name, context, requires,
                                             fn, watch_func))
        return steps

    def __repr__(self):
        """Object represented as a string."""
        return "<CFNgin.plan.Step:%s>" % (self.stack.name,)

    def __str__(self):
        """Object displayed as a string."""
        return self.stack.name


class Graph(object):
    """Graph represents a graph of steps.

    The :class:`Graph` helps organize the steps needed to execute a particular
    action for a set of :class:`runway.cfngin.stack.Stack` objects. When
    initialized with a set of steps, it will first build a Directed Acyclic
    Graph from the steps and their dependencies.

    Attributes:
        dag (:class:`runway.cfngin.dag.DAG`): an optional
            :class:`runway.cfngin.dag.DAG` object. If one is not provided, a
            new one will be initialized.
        steps (Dict[str, :class:`Step`]): Dict with key of step name and
            value of :class:`Step`.

    Example:

    >>> dag = DAG()
    >>> a = Step("a", fn=build)
    >>> b = Step("b", fn=build)
    >>> dag.add_step(a)
    >>> dag.add_step(b)
    >>> dag.connect(a, b)

    """

    def __init__(self, steps=None, dag=None):
        """Instantiate class.

        Args:
            steps (Optional[Dict[str, :class:`Step`]]): Dict with key of step
                name and value of :class:`Step` for steps to initialize the
                Graph with. Note that if this is provided, a pre-configured
                :class:`runway.cfngin.dag.DAG` that already includes these
                steps should also be provided..
            dag (Optional[:class:`runway.cfngin.dag.DAG`]): An optional
                :class:`runway.cfngin.dag.DAG` object. If one is not provided,
                a new one will be initialized.

        """
        self.steps = steps or {}
        self.dag = dag or DAG()

    def add_step(self, step, add_dependencies=False, add_dependants=False):
        """Add a step to the graph.

        Args:
            step (:class:`Step`): The step to be added.
            add_dependencies (bool): Connect steps that need to be completed
                before this step.
            add_dependants (bool): Connect steps that require this step.

        """
        self.steps[step.name] = step
        self.dag.add_node(step.name)

        if add_dependencies:
            for dep in step.requires:
                self.connect(step.name, dep)

        if add_dependants:
            for parent in step.required_by:
                self.connect(parent, step.name)

    def add_step_if_not_exists(self, step, add_dependencies=False,
                               add_dependants=False):
        """Try to add a step to the graph.

        Can be used when failure to add is acceptable.

        Args:
            step (:class:`Step`): The step to be added.
            add_dependencies (bool): Connect steps that need to be completed
                before this step.
            add_dependants (bool): Connect steps that require this step.

        """
        if self.steps.get(step.name):
            return

        self.steps[step.name] = step
        self.dag.add_node_if_not_exists(step.name)

        if add_dependencies:
            for dep in step.requires:
                try:
                    self.connect(step.name, dep)
                except GraphError:
                    continue

        if add_dependants:
            for parent in step.required_by:
                try:
                    self.connect(parent, step.name)
                except GraphError:
                    continue

    def add_steps(self, steps):
        """Add a list of steps.

        Args:
            steps (List[:class:`Step`]): The step to be added.

        """
        for step in steps:
            self.add_step(step)

        for step in steps:
            for dep in step.requires:
                self.connect(step.name, dep)

            for parent in step.required_by:
                self.connect(parent, step.name)

    def pop(self, step, default=None):
        """Remove a step from the graph.

        Args:
            step (:class:`Step`): The step to remove from the graph.
            default (Any): Returned if the step could not be popped

        Returns:
            Any

        """
        self.dag.delete_node_if_exists(step.name)
        return self.steps.pop(step.name, default)

    def connect(self, step, dep):
        """Connect a dependency to a step.

        Args:
            step (str): Step name to add a dependency to.
            dep (str): Name of dependent step.

        """
        try:
            self.dag.add_edge(step, dep)
        except KeyError as err:
            raise GraphError(err, step, dep)
        except DAGValidationError as err:
            raise GraphError(err, step, dep)

    def transitive_reduction(self):
        """Perform a transitive reduction on the underlying DAG.

        The transitive reduction of a graph is a graph with as few edges as
        possible with the same reachability as the original graph.

        See https://en.wikipedia.org/wiki/Transitive_reduction

        """
        self.dag.transitive_reduction()

    def walk(self, walker, walk_func):
        """Walk the steps of the graph.

        Args:
            walker (Callable[[:class:`runway.cfngin.dag.DAG`], Any]): Function
                used to walk the steps.
            walk_func (Callable[[:class:`Step`], Any]): Function called with a
                :class:`Step` as the only argument for each step of the plan.

        """
        def fn(step_name):
            """Get a step by step name and execute the ``walk_func`` on it.

            Args:
                step_name (str): Name of a step.

            """
            step = self.steps[step_name]
            return walk_func(step)

        return walker(self.dag, fn)

    def downstream(self, step_name):
        """Return the direct dependencies of the given step."""
        return list(self.steps[dep] for dep in self.dag.downstream(step_name))

    def transposed(self):
        """Return a "transposed" version of this graph.

        Useful for walking in reverse.

        """
        return Graph(steps=self.steps, dag=self.dag.transpose())

    def filtered(self, step_names):
        """Return a "filtered" version of this graph.

        Args:
            step_names (List[str]): Steps to filter.
        """
        return Graph(steps=self.steps, dag=self.dag.filter(step_names))

    def topological_sort(self):
        """Perform a topological sort of the underlying DAG.

        Returns:
            List[Step]

        """
        nodes = self.dag.topological_sort()
        return [self.steps[step_name] for step_name in nodes]

    def to_dict(self):
        """Return the underlying DAG as a dictionary."""
        return self.dag.graph

    def dumps(self, indent=None):
        """Output the graph as a json seralized string for storage.

        Args:
            indent (Optional[int]): Number of spaces for each indentation.

        Returns:
            str

        """
        return json.dumps(self.to_dict(), default=json_serial, indent=indent)

    @classmethod
    def from_dict(cls, graph_dict, context):
        """Create a Graph from a graph dict.

        Args:
            graph_dict (Dict[str, List[str]]): The dictionary used to
                create the graph.
            context (:class:`runway.cfngin.context.Context`): Required to init
                stacks.

        Returns:
            :class:`Graph`

        """
        return cls.from_steps(Step.from_persistent_graph(graph_dict, context))

    @classmethod
    def from_steps(cls, steps):
        """Create a Graph from Steps.

        Args:
            steps (List[:class:`Step`]): Steps used to create the graph.

        Returns:
            :class:`Graph`

        """
        graph = cls()
        graph.add_steps(steps)
        return graph

    def __str__(self):
        """Object displayed as a string."""
        return self.dumps()


class Plan(object):
    """A convenience class for working on a Graph.

    Attributes:
        context (:class:`runway.cfngin.context.Context`): Context object.
        description (str): Plan description.
        graph (Graph): Graph of the plan.
        id (str): UUID for the plan.
        reverse (bool): The graph has been transposed for walking in reverse.
        require_unlocked (bool): Require the persistent graph to be unlocked
            before executing steps.

    """

    def __init__(self, description, graph, context=None,
                 reverse=False, require_unlocked=True):
        """Initialize class.

        Args:
            description (str): Description of what the plan is going to do.
            graph (:class:`Graph`): Local graph used for the plan.
            context (:class:`runway.cfngin.context.Context`): Context object.
            reverse (bool): Transpose the graph for walking in reverse.
            require_unlocked (bool): Require the persistent graph to be
                unlocked before executing steps.

        """
        self.context = context
        self.description = description
        self.id = uuid.uuid4()
        self.reverse = reverse
        self.require_unlocked = require_unlocked

        if self.reverse:
            graph = graph.transposed()

        if self.context:
            self.locked = self.context.persistent_graph_locked

            if self.context.stack_names:
                nodes = []
                for target in self.context.stack_names:
                    if graph.steps.get(target):
                        nodes.append(target)
                graph = graph.filtered(nodes)
        else:
            self.locked = False

        self.graph = graph

    def outline(self, level=logging.INFO, message=""):
        """Print an outline of the actions the plan is going to take.

        The outline will represent the rough ordering of the steps that will be
        taken.

        Args:
            level (Optional[int]): a valid log level that should be used to log
                the outline
            message (Optional[str]): a message that will be logged to
                the user after the outline has been logged.

        """
        steps = 1
        LOGGER.log(level, "Plan \"%s\":", self.description)
        for step in self.steps:
            LOGGER.log(
                level,
                "  - step: %s: target: \"%s\", action: \"%s\"",
                steps,
                step.name,
                step.fn.__name__,
            )
            steps += 1

        if message:
            LOGGER.log(level, message)

    def dump(self, directory, context, provider=None):
        """Output the rendered blueprint for all stacks in the plan.

        Args:
            directory (str): Directory where files will be created.
            context (:class:`runway.cfngin.context.Contest`): Current CFNgin
                context.
            provider (:class:`runway.cfngin.providers.aws.default.Provider`):
                Provider to use when resolving the blueprints.

        """
        LOGGER.info("Dumping \"%s\"...", self.description)
        directory = os.path.expanduser(directory)
        if not os.path.exists(directory):
            os.makedirs(directory)

        def walk_func(step):
            """Walk function."""
            step.stack.resolve(
                context=context,
                provider=provider,
            )
            blueprint = step.stack.blueprint
            filename = stack_template_key_name(blueprint)
            path = os.path.join(directory, filename)

            blueprint_dir = os.path.dirname(path)
            if not os.path.exists(blueprint_dir):
                os.makedirs(blueprint_dir)

            LOGGER.info("Writing stack \"%s\" -> %s", step.name, path)
            with open(path, "w") as _file:
                _file.write(blueprint.rendered)

            return True

        return self.graph.walk(walk, walk_func)

    def execute(self, *args, **kwargs):
        """Walk each step in the underlying graph.

        Raises:
            PersistentGraphLocked: Raised if the persistent graph is
                locked prior to execution and this session did not lock it.
            PlanFailed: Raised if any of the steps fail.

        """
        if self.locked and self.require_unlocked:
            raise PersistentGraphLocked
        self.walk(*args, **kwargs)

        failed_steps = [step for step in self.steps if step.status == FAILED]
        if failed_steps:
            raise PlanFailed(failed_steps)

    def walk(self, walker):
        """Walk each step in the underlying graph, in topological order.

        Args:
            walker (func): a walker function to be passed to
                :class:`runway.cfngin.dag.DAG` to walk the graph.

        """
        def walk_func(step):
            """Execute a :class:`Step` wile walking the graph.

            Handles updating the persistent graph if one is being used.

            Args:
                step (:class:`Step`): :class:`Step` to execute.

            Returns:
                bool

            """
            # Before we execute the step, we need to ensure that it's
            # transitive dependencies are all in an "ok" state. If not, we
            # won't execute this step.
            for dep in self.graph.downstream(step.name):
                if not dep.ok:
                    step.set_status(FailedStatus("dependency has failed"))
                    return step.ok

            result = step.run()

            if not self.context or not self.context.persistent_graph:
                return result

            if (step.completed or
                    (step.skipped and
                     step.status.reason == ('does not exist in '
                                            'cloudformation'))):
                if step.fn.__name__ == '_destroy_stack':
                    self.context.persistent_graph.pop(step)
                    LOGGER.debug("Removed step '%s' from the persistent graph",
                                 step.name)
                elif step.fn.__name__ == '_launch_stack':
                    self.context.persistent_graph.add_step_if_not_exists(
                        step, add_dependencies=True, add_dependants=True
                    )
                    LOGGER.debug("Added step '%s' to the persistent graph",
                                 step.name)
                else:
                    return result
                self.context.put_persistent_graph(self.lock_code)
            return result

        return self.graph.walk(walker, walk_func)

    @property
    def lock_code(self):
        """Code to lock/unlock the persistent graph.

        Returns:
            str

        """
        return str(self.id)

    @property
    def steps(self):
        """Return a list of all steps in the plan.

        Returns:
            List[:class:`Step`]

        """
        steps = self.graph.topological_sort()
        steps.reverse()
        return steps

    @property
    def step_names(self):
        """Return a list of all step names.

        Returns:
            List[str]

        """
        return [step.name for step in self.steps]

    def keys(self):
        """Return a list of all step names.

        Returns:
            List[str]

        """
        return self.step_names



"""Tests for terraform module."""
import unittest

from runway.module.terraform import update_env_vars_with_tf_var_values


class TerraformFunctionTester(unittest.TestCase):
    """Test Terraform module support functions."""

    def test_update_env_vars_with_tf_var_values(self):
        """Test update_env_vars_with_tf_var_values."""
        env_vars = update_env_vars_with_tf_var_values(
            {},
            {'foo': 'bar',
             'list': ['test1', 'test2', 'test3'],
             'map': {'one': 'two',
                     'three': 'four'}}
        )

        # This can be simplified post-python2 (with dict order preservation)
        # self.assertDictEqual(
        #     env_vars,
        #     {'TF_VAR_foo': 'bar',
        #     {'TF_VAR_list': '[test1,test2,test3]',
        #      'TF_VAR_map': '{ one = "two", three = "four" }'}
        # )
        self.assertTrue(env_vars['TF_VAR_foo'] == 'bar')
        self.assertTrue(env_vars['TF_VAR_list'] == '[test1,test2,test3]')
        self.assertRegexpMatches(env_vars['TF_VAR_map'], r'one = "two"')  # noqa pylint: disable=deprecated-method
        self.assertRegexpMatches(env_vars['TF_VAR_map'], r'three = "four"')  # noqa pylint: disable=deprecated-method



"""Tests for runway.cfngin.tokenize_userdata."""
import unittest

import yaml

from runway.cfngin.tokenize_userdata import cf_tokenize


class TestCfTokenize(unittest.TestCase):
    """Tests for runway.cfngin.tokenize_userdata."""

    def test_tokenize(self):
        """Test tokenize."""
        user_data = [
            "field0",
            "Ref(\"SshKey\")",
            "field1",
            "Fn::GetAtt(\"Blah\", \"Woot\")"
        ]
        user_data_dump = yaml.dump(user_data)
        parts = cf_tokenize(user_data_dump)
        self.assertIsInstance(parts[1], dict)
        self.assertIsInstance(parts[3], dict)
        self.assertEqual(parts[1]["Ref"], "SshKey")
        self.assertEqual(parts[3]["Fn::GetAtt"], ["Blah", "Woot"])
        self.assertEqual(len(parts), 5)



"""Testing remote Path Sources in deployments."""
import os
import glob

from send2trash import send2trash

from integration_tests.integration_test import IntegrationTest
from integration_tests.util import copy_file, import_tests, execute_tests


class TestSources(IntegrationTest):
    """Test sources."""

    base_dir = os.path.abspath(os.path.dirname(__file__))
    fixtures_dir = os.path.join(base_dir, 'fixtures')
    tests_dir = os.path.join(base_dir, 'tests')

    sources_test_dir = os.path.join(base_dir, 'sources_test')

    def __init__(self, logger):
        """Initialize."""
        IntegrationTest.__init__(self, logger)

    def copy_runway(self, template):
        """Copy runway template to proper directory."""
        template_file = os.path.join(self.fixtures_dir, 'runway-{}.yml'.format(template))
        copy_file(template_file, os.path.join(self.sources_test_dir, 'runway.yml'))

    def run(self):
        """Find all tests and run them."""
        import_tests(self.logger, self.tests_dir, 'test_*')
        tests = [test(self.logger) for test in TestSources.__subclasses__()]
        self.logger.debug('FOUND TESTS: %s', tests)
        self.set_environment('dev')
        err_count = execute_tests(tests, self.logger)
        assert err_count == 0
        return err_count

    def clean(self):
        """Clean up fixture test directory."""
        file_types = ('*.yaml', '*.yml')
        templates = []
        for file_type in file_types:
            templates.extend(glob.glob(os.path.join(
                self.sources_test_dir,
                file_type
            )))
        for template in templates:
            if os.path.isfile(template):
                self.logger.debug('send2trash: "%s"', template)
                send2trash(template)
        folders = ['.runway_cache']
        for folder in folders:
            folder_path = os.path.join(self.sources_test_dir, folder)
            if os.path.isdir(folder_path):
                self.logger.debug('send2trash: "%s"', folder_path)
                send2trash(folder_path)

    def teardown(self):
        """Teardown resources create during init."""
        self.clean()



"""Environment variable lookup."""
# pylint: disable=unused-argument,arguments-differ
import os

from runway.lookups.handlers.base import LookupHandler

from ...util import read_value_from_path

TYPE_NAME = "envvar"


class EnvvarLookup(LookupHandler):
    """Environment variable lookup."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Retrieve an environment variable.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        Example:
            ::

                # With CFNgin we would reference the environment variable like this:
                conf_key: ${envvar ENV_VAR_NAME}

            You can optionally store the value in a file, ie::

                $ cat envvar_value.txt
                ENV_VAR_NAME

            and reference it within CFNgin (NOTE: the path should be relative
            to the CFNgin config file)::

                conf_key: ${envvar file://envvar_value.txt}

                # Both of the above would resolve to
                conf_key: ENV_VALUE

        """
        value = read_value_from_path(value)

        try:
            return os.environ[value]
        except KeyError:
            raise ValueError('EnvVar "{}" does not exist'.format(value))



"""Tests for Terraform."""
import glob
import os

from send2trash import send2trash

from integration_tests.integration_test import IntegrationTest
from integration_tests.util import copy_file, execute_tests, import_tests


class Terraform(IntegrationTest):
    """Base class for all Terraform testing."""

    base_dir = os.path.abspath(os.path.dirname(__file__))
    template_dir = os.path.join(base_dir, 'templates')

    tf_test_dir = os.path.join(base_dir, 'terraform_test.tf')
    tf_state_dir = os.path.join(base_dir, 'tf_state.cfn')

    def copy_runway(self, template):
        """Copy runway template to proper directory."""
        template_file = os.path.join(self.template_dir, 'runway-{}.yml'.format(template))
        copy_file(template_file, os.path.join(self.base_dir, 'runway.yml'))

    def copy_template(self, template, name='main.tf'):
        """Copy template to Terraform module folder."""
        template_file = os.path.join(self.template_dir, template)
        copy_file(template_file, os.path.join(self.tf_test_dir, name))

    def clean(self):
        """Clean up Terraform module directory."""
        file_types = ('*.tf', '.terraform-version', '*.yml', 'local_backend')
        templates = []
        for file_type in file_types:
            templates.extend(glob.glob(os.path.join(self.tf_test_dir, file_type)))
            templates.extend(glob.glob(os.path.join(self.base_dir, file_type)))

        for template in templates:
            if os.path.isfile(template):
                self.logger.debug('send2trash: "%s"', template)
                send2trash(template)
        folders = ('.terraform', 'terraform.tfstate.d')
        for folder in folders:
            folder_path = os.path.join(self.tf_test_dir, folder)
            if os.path.isdir(folder_path):
                self.logger.debug('send2trash: "%s"', folder_path)
                send2trash(folder_path)

    def set_tf_version(self, version=11):
        """Copy version file to module directory."""
        version_file = 'tf-v{}.version'.format(version)
        self.copy_template(version_file, '.terraform-version')

    def run(self):
        """Find all Terraform tests and run them."""
        import_tests(self.logger, self.tests_dir, 'test_*')
        self.set_environment('dev')
        self.set_env_var('CI', '1')
        tests = [test(self.logger, self.environment)
                 for test in Terraform.__subclasses__()]
        if not tests:
            raise Exception('No tests were found.')
        self.logger.debug('FOUND TESTS: %s', tests)
        err_count = execute_tests(tests, self.logger)
        assert err_count == 0  # assert that all subtests were successful
        return err_count

    def teardown(self):
        """Teardown resources create during init."""
        self.clean()



"""DynamoDB lookup."""
# pylint: disable=arguments-differ,unused-argument
import re

from botocore.exceptions import ClientError

from runway.lookups.handlers.base import LookupHandler

from ...session_cache import get_session
from ...util import read_value_from_path

TYPE_NAME = 'dynamodb'


class DynamodbLookup(LookupHandler):
    """DynamoDB lookup."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Get a value from a DynamoDB table.

        Args:
            value (str): Parameter(s) given to this lookup.
                ``[<region>:]<tablename>@<primarypartionkey>:<keyvalue>.<keyvalue>...``
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        .. note:: The region is optional, and defaults to the environment's
                  ``AWS_DEFAULT_REGION`` if not specified.

        """
        value = read_value_from_path(value)
        table_info = None
        table_keys = None
        region = None
        table_name = None
        if '@' in value:
            table_info, table_keys = value.split('@', 1)
            if ':' in table_info:
                region, table_name = table_info.split(':', 1)
            else:
                table_name = table_info
        else:
            raise ValueError('Please make sure to include a tablename')

        if not table_name:
            raise ValueError('Please make sure to include a DynamoDB table '
                             'name')

        table_lookup, table_keys = table_keys.split(':', 1)

        table_keys = table_keys.split('.')

        key_dict = _lookup_key_parse(table_keys)
        new_keys = key_dict['new_keys']
        clean_table_keys = key_dict['clean_table_keys']

        projection_expression = _build_projection_expression(clean_table_keys)

        # lookup the data from DynamoDB
        dynamodb = get_session(region).client('dynamodb')
        try:
            response = dynamodb.get_item(
                TableName=table_name,
                Key={
                    table_lookup: new_keys[0]
                },
                ProjectionExpression=projection_expression
            )
        except ClientError as err:
            if err.response['Error']['Code'] == 'ResourceNotFoundException':
                raise ValueError(
                    'Cannot find the DynamoDB table: {}'.format(table_name))
            if err.response['Error']['Code'] == 'ValidationException':
                raise ValueError(
                    'No DynamoDB record matched the partition key: '
                    '{}'.format(table_lookup))
            raise ValueError('The DynamoDB lookup {} had an error: '
                             '{}'.format(value, err))
        # find and return the key from the dynamo data returned
        if 'Item' in response:
            return _get_val_from_ddb_data(response['Item'], new_keys[1:])
        raise ValueError(
            'The DynamoDB record could not be found using the following '
            'key: {}'.format(new_keys[0]))


def _lookup_key_parse(table_keys):
    """Return the order in which the stacks should be executed.

    Args:
        dependencies (dict): a dictionary where each key should be the
            fully qualified name of a stack whose value is an array of
            fully qualified stack names that the stack depends on. This is
            used to generate the order in which the stacks should be
            executed.

    Returns:
        dict: includes a dict of lookup types with data types ('new_keys')
              and a list of the lookups with without ('clean_table_keys')

    """
    # we need to parse the key lookup passed in
    regex_matcher = r'\[([^\]]+)]'
    valid_dynamodb_datatypes = ['M', 'S', 'N', 'L']
    clean_table_keys = []
    new_keys = []

    for key in table_keys:
        match = re.search(regex_matcher, key)
        if match:
            # the datatypes are pulled from the dynamodb docs
            if match.group(1) in valid_dynamodb_datatypes:
                match_val = str(match.group(1))
                key = key.replace(match.group(0), '')
                new_keys.append({match_val: key})
                clean_table_keys.append(key)
            else:
                raise ValueError(
                    ('CFNgin does not support looking up the datatype: {}')
                    .format(str(match.group(1))))
        else:
            new_keys.append({'S': key})
            clean_table_keys.append(key)
    key_dict = {}
    key_dict['new_keys'] = new_keys
    key_dict['clean_table_keys'] = clean_table_keys

    return key_dict


def _build_projection_expression(clean_table_keys):
    """Return a projection expression for the DynamoDB lookup.

    Args:
        clean_table_keys (Dict[str, Any]): Keys without the data types attached.

    Returns:
        str: A projection expression for the DynamoDB lookup.

    """
    projection_expression = ''
    for key in clean_table_keys[:-1]:
        projection_expression += ('{},').format(key)
    projection_expression += clean_table_keys[-1]
    return projection_expression


def _get_val_from_ddb_data(data, keylist):
    """Return the value of the lookup.

    Args:
        data (Dict[str, Any]): The raw DynamoDB data.
        keylist(List[Dict[str, str]]): A list of keys to lookup. This must
            include the datatype.

    Returns:
        Any: It returns the value from the DynamoDB record, and casts it
            to a matching python datatype.

    """
    next_type = None
    # iterate through the keylist to find the matching key/datatype
    for key in keylist:
        for k in key:
            if next_type is None:
                data = data[key[k]]
            else:
                temp_dict = data[next_type]
                data = temp_dict[key[k]]
            next_type = k
    if next_type == 'L':
        # if type is list, convert it to a list and return
        return _convert_ddb_list_to_list(data[next_type])
    if next_type == 'N':
        # TODO: handle various types of 'number' datatypes, (e.g. int, double)
        # if a number, convert to an int and return
        return int(data[next_type])
    # else, just assume its a string and return
    return str(data[next_type])


def _convert_ddb_list_to_list(conversion_list):
    """Return a python list without the DynamoDB datatypes.

    Args:
        conversion_list (Dict[str, Any]): A DynamoDB list which includes the
            datatypes.

    Returns:
        List[Any]: Returns A sanitized list without the datatypes.

    """
    ret_list = []
    for val in conversion_list:
        for v in val:
            ret_list.append(val[v])
    return ret_list



"""Command hook."""
# pylint: disable=unused-argument
import logging
import os
from subprocess import PIPE, Popen

from ..exceptions import ImproperlyConfigured

LOGGER = logging.getLogger(__name__)


def _devnull():
    return open(os.devnull, 'wb')


def run_command(provider, context, command, capture=False, interactive=False,
                ignore_status=False, quiet=False, stdin=None, env=None,
                **kwargs):
    """Run a custom command as a hook.

    Args:
        provider (:class:`runway.cfngin.providers.base.BaseProvider`): Provider
            instance. (passed in by CFNgin)
        context (:class:`runway.cfngin.context.Context`): Context instance.
            (passed in by CFNgin)

    Keyword Args:
        command (Union[str, List[str]]): Command(s) to run.
        capture (bool): If enabled, capture the command's stdout and stderr,
            and return them in the hook result. (*default:* ``False``)
        interactive (bool): If enabled, allow the command to interact with
            stdin. Otherwise, stdin will be set to the null device.
            (*default:* ``False``)
        ignore_status (bool): Don't fail the hook if the command returns a
            non-zero status. (*default:* ``False``)
        quiet (bool): Redirect the command's stdout and stderr to the null
            device, silencing all output. Should not be enabled if
            ``capture`` is also enabled. (*default:* ``False``)
        stdin (Optional[str]): String to send to the stdin of the command.
            Implicitly disables ``interactive``.
        env (Optional[Dict[str, str]]): Dictionary of environment variable
            overrides for the command context. Will be merged with the current
            environment.
        **kwargs (Any): Any other arguments will be forwarded to the
            ``subprocess.Popen`` function. Interesting ones include: ``cwd``
            and ``shell``.

    Examples:
        .. code-block:: yaml

            pre_build:
              command_copy_environment:
                path: runway.cfngin.hooks.command.run_command
                required: true
                enabled: true
                data_key: copy_env
                args:
                  command: ['cp', 'environment.template', 'environment']
              command_git_rev_parse:
                path: runway.cfngin.hooks.command.run_command
                required: true
                enabled: true
                data_key: get_git_commit
                args:
                  command: ['git', 'rev-parse', 'HEAD']
                  cwd: ./my-git-repo
                  capture: true
              command_npm_install:
                path: runway.cfngin.hooks.command.run_command
                args:
                  command: '`cd $PROJECT_DIR/project; npm install`'
                  env:
                    PROJECT_DIR: ./my-project
                    shell: true

    """
    if quiet and capture:
        raise ImproperlyConfigured(
            __name__ + '.run_command',
            'Cannot enable `quiet` and `capture` options simultaneously')

    if quiet:
        out_err_type = _devnull()
    elif capture:
        out_err_type = PIPE
    else:
        out_err_type = None

    if interactive:
        in_type = None
    elif stdin:
        in_type = PIPE
    else:
        in_type = _devnull()

    if env:
        full_env = os.environ.copy()
        full_env.update(env)
        env = full_env

    LOGGER.info('Running command: %s', command)

    proc = Popen(command, stdin=in_type, stdout=out_err_type,
                 stderr=out_err_type, env=env, **kwargs)
    try:
        out, err = proc.communicate(stdin)
        status = proc.wait()

        if status == 0 or ignore_status:
            return {
                'returncode': proc.returncode,
                'stdout': out,
                'stderr': err
            }

        # Don't print the command line again if we already did earlier
        if LOGGER.isEnabledFor(logging.INFO):
            LOGGER.warning('Command failed with returncode %d', status)
        else:
            LOGGER.warning('Command failed with returncode %d: %s', status,
                           command)

        return None
    finally:
        if proc.returncode is None:
            proc.kill()
